import abc

from six import add_metaclass

from hackle.decision import DecisionReason
from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.evaluation.evaluator.experiment.experiment_evaluation import ExperimentEvaluation
from hackle.internal.evaluation.evaluator.experiment.experiment_request import ExperimentRequest
from hackle.internal.evaluation.match.condition_matcher import ConditionMatcher
from hackle.internal.evaluation.match.value_operator_matcher import ValueOperatorMatcher
from hackle.internal.model.entities import TargetCondition, Experiment
from hackle.internal.type import hackle_types


class ExperimentConditionMatcher(ConditionMatcher):

    def __init__(self, ab_test_matcher, feature_flag_matcher):
        """
        :param AbTestConditionMatcher ab_test_matcher:
        :param FeatureFlagConditionMatcher feature_flag_matcher:
        """
        self.__ab_test_matcher = ab_test_matcher
        self.__feature_flag_matcher = feature_flag_matcher

    def matches(self, request, context, condition):
        if condition.key.type == 'AB_TEST':
            return self.__ab_test_matcher.matches(request, context, condition)
        elif condition.key.type == 'FEATURE_FLAG':
            return self.__feature_flag_matcher.matches(request, context, condition)
        else:
            raise Exception('Unsupported TargetKeyType[{}]'.format(condition.key.type))


@add_metaclass(abc.ABCMeta)
class ExperimentMatcher(object):
    def __init__(self, evaluator):
        """
        :param Evaluator evaluator:
        """
        self.__evaluator = evaluator

    def matches(self, request, context, condition):
        """
        :param Evaluator.Request request:
        :param Evaluator.Context context:
        :param TargetCondition condition:

        :rtype: bool
        """
        key = hackle_types.as_int_or_none(condition.key.name)
        if key is None:
            raise Exception('Invalid key [{}, {}]'.format(condition.key.type, condition.key.name))
        experiment = self._experiment_or_none(request, key)
        if experiment is None:
            return False

        evaluation = context.get_evaluation_or_none(experiment)
        if evaluation is None:
            evaluation = self.__evaluate(request, context, experiment)
        return self._evaluation_matches(evaluation, condition)

    def __evaluate(self, request, context, experiment):
        """
        :param Evaluator.Request request:
        :param Evaluator.Context context:
        :param Experiment experiment:

        :rtype: ExperimentEvaluation
        """
        experiment_request = ExperimentRequest.from_request(request, experiment)
        evaluation = self.__evaluator.evaluate(experiment_request, context)
        if not isinstance(evaluation, ExperimentEvaluation):
            raise Exception('Unexpected evaluation: {} (expected: ExperimentEvaluation)'.format(evaluation))
        experiment_evaluation = self._resolve(request, evaluation)
        context.add_evaluation(experiment_evaluation)
        return experiment_evaluation

    @abc.abstractmethod
    def _experiment_or_none(self, request, key):
        """
        :param Evaluator.Request request:
        :param int key:

        :rtype: Experiment or None
        """
        pass

    @abc.abstractmethod
    def _resolve(self, request, evaluation):
        """
        :param Evaluator.Request request:
        :param ExperimentEvaluation evaluation:

        :rtype: ExperimentEvaluation
        """
        pass

    @abc.abstractmethod
    def _evaluation_matches(self, evaluation, condition):
        """
        :param ExperimentEvaluation evaluation:
        :param TargetCondition condition:

        :rtype: bool
        """
        pass


class AbTestConditionMatcher(ExperimentMatcher):
    __AB_TEST_MATCHED_REASONS = [
        DecisionReason.OVERRIDDEN,
        DecisionReason.TRAFFIC_ALLOCATED,
        DecisionReason.TRAFFIC_ALLOCATED_BY_TARGETING,
        DecisionReason.EXPERIMENT_COMPLETED,
    ]

    def __init__(self, evaluator, value_operator_matcher):
        """
        :param Evaluator evaluator:
        :param ValueOperatorMatcher value_operator_matcher:
        """
        super(AbTestConditionMatcher, self).__init__(evaluator)
        self.__value_operator_matcher = value_operator_matcher

    def _experiment_or_none(self, request, key):
        return request.workspace.get_experiment_or_none(key)

    def _resolve(self, request, evaluation):
        if isinstance(request, ExperimentRequest) and evaluation.reason == DecisionReason.TRAFFIC_ALLOCATED:
            return evaluation.copy_with(DecisionReason.TRAFFIC_ALLOCATED_BY_TARGETING)
        return evaluation

    def _evaluation_matches(self, evaluation, condition):
        if not evaluation.reason in self.__AB_TEST_MATCHED_REASONS:
            return False
        return self.__value_operator_matcher.matches(evaluation.variation_key, condition.match)


class FeatureFlagConditionMatcher(ExperimentMatcher):
    def __init__(self, evaluator, value_operator_matcher):
        """
        :param Evaluator evaluator:
        :param ValueOperatorMatcher value_operator_matcher:
        """
        super(FeatureFlagConditionMatcher, self).__init__(evaluator)
        self.__value_operator_matcher = value_operator_matcher

    def _experiment_or_none(self, request, key):
        return request.workspace.get_feature_flag_or_none(key)

    def _resolve(self, request, evaluation):
        return evaluation

    def _evaluation_matches(self, evaluation, condition):
        on = evaluation.variation_key != 'A'
        return self.__value_operator_matcher.matches(on, condition.match)



from unittest import TestCase

from hackle.internal.user.identifiers_builder import IdentifiersBuilder


class IdentifierBuilderTest(TestCase):

    def test_add_identifier_type이_None이면_추가하지_않는다(self):
        self.assertEqual({}, IdentifiersBuilder().add(None, 'id').build())

    def test_add_identifier_type이_str_타입이_아니면_추가하지_않는다(self):
        self.assertEqual({}, IdentifiersBuilder().add(42, 'id').build())

    def test_add_identifier_type이_128자보다_크면_추가하지_않는다(self):
        self.assertEqual({}, IdentifiersBuilder().add('a' * 129, 'id').build())

    def test_add_identifier_value가_None이면_추가하지_않는다(self):
        self.assertEqual({}, IdentifiersBuilder().add('my_id', None).build())

    def test_add_identifier_value가_512자보다_크면_추가하지_않는다(self):
        self.assertEqual({}, IdentifiersBuilder().add('my_id', 'a' * 513).build())

    def test_add(self):
        builder = IdentifiersBuilder()
        builder.add('id3', 'a')
        builder.add('id2', 'myId')
        builder.add('id1', 'hello')
        builder.add('a' * 128, 'a' * 512)
        builder.add('num1', 1)
        builder.add('invalid2', None)
        identifier = builder.build()

        self.assertEqual({
            'id1': 'hello',
            'id2': 'myId',
            'id3': 'a',
            'a' * 128: 'a' * 512,
            'num1': '1'
        }, identifier)

    def test_add_identifiers_None이면_추가하지_않는다(self):
        self.assertEqual({}, IdentifiersBuilder().add_identifiers(None).build())

    def test_add_identifiers_dict가_아니면_추가하지_않는다(self):
        self.assertEqual({}, IdentifiersBuilder().add_identifiers('must_be_dict').build())

    def test_add_identifiers(self):
        identifiers = {
            'id1': 'hello',
            'id2': 'myId',
            'id3': 'a',
            'a' * 128: 'a' * 512
        }

        builder = IdentifiersBuilder()
        builder.add_identifiers(identifiers)
        builder.add('bbb', 'ccc')
        actual = builder.build()

        self.assertEqual({
            'id1': 'hello',
            'id2': 'myId',
            'id3': 'a',
            'a' * 128: 'a' * 512,
            'bbb': 'ccc'
        }, actual)



from threading import Lock


class AtomicInteger(object):

    def __init__(self, value=0):
        self.__value = value
        self.__lock = Lock()

    def get(self):
        """
        :rtype: int
        """
        with self.__lock:
            return self.__value

    def add_and_get(self, delta):
        """
        :param int delta:
        :rtype: int
        """
        with self.__lock:
            self.__value += delta
            return self.__value

    def get_and_add(self, delta):
        """
        :param int delta:
        :rtype: int
        """
        with self.__lock:
            old_value = self.__value
            self.__value += delta
            return old_value



import time
from threading import Thread, Event

from hackle.internal.concurrent.schedule.scheduler import Scheduler, ScheduledJob
from hackle.internal.logger.log import Log
from hackle.internal.time.time_unit import SECONDS


class ThreadScheduledJob(ScheduledJob):

    def __init__(self, delay, period, task):
        super(ThreadScheduledJob, self).__init__()
        self.__delay = delay
        self.__period = period
        self.__task = task
        self.__stop = Event()
        self.__thread = Thread(target=self.__run)
        self.__thread.daemon = True

    def start(self):
        self.__thread.start()

    def cancel(self):
        self.__stop.set()

    def __run(self):
        if self.__delay > 0:
            if self.__stop.wait(self.__delay):
                return

        is_stopped = self.__stop.is_set()
        while not is_stopped:
            next_time = time.time() + self.__period
            try:
                self.__task()
            except Exception as e:
                Log.get().error("Unexpected exception on scheduler: {}".format(str(e)))
            delay = next_time - time.time()
            is_stopped = self.__stop.wait(delay) if delay > 0 else self.__stop.is_set()


class ThreadScheduler(Scheduler):

    def __init__(self):
        super(ThreadScheduler, self).__init__()

    def schedule_periodically(self, delay, period, unit, task):
        job = ThreadScheduledJob(SECONDS.convert(delay, unit), SECONDS.convert(period, unit), task)
        job.start()
        return job

    def close(self):
        pass



from unittest import mock
from unittest.mock import Mock

from tests import base
from hackle.internal.model.entities import Slot, Bucket
from hackle.internal.evaluation.bucket.bucketer import Bucketer


class BucketerTest(base.BaseTest):

    def setUp(self):
        self.sut = Bucketer()

    def test_버켓에서_계산된_슬롯번호로_슬롯을_가져온다(self):
        # given
        slot = Mock(spec=Slot)
        bucket = Mock(spec=Bucket)
        bucket.get_slot_or_none.return_value = slot
        bucket.seed = 123
        bucket.slot_size = 10000

        with mock.patch.object(self.sut, '_calculate_slot_number',
                               wraps=self.sut._calculate_slot_number) as _calculate_slot_number:
            _calculate_slot_number.return_value = 320

            # when
            actual = self.sut.bucketing(bucket, 'test_id')

            # then
            bucket.get_slot_or_none.assert_called_once_with(320)
            _calculate_slot_number.assert_called_once_with('test_id', 123, 10000)
            assert actual == slot



import unittest
from unittest.mock import Mock

from hackle.internal.evaluation.bucket.bucketer import Bucketer
from hackle.internal.evaluation.container.container_resolver import ContainerResolver
from hackle.internal.model.entities import Container, Bucket, Slot, ContainerGroup
from hackle.internal.user.internal_hackle_user import InternalHackleUser
from hackle.internal.workspace.workspace import Workspace
from tests.internal.evaluation.evaluator.experiment.test_experiment_request import create_experiment_request
from tests.internal.model.test_entities import create_experiment


class ContainerResolverTest(unittest.TestCase):

    def setUp(self):
        self.bucketer = Mock(spec=Bucketer)
        self.sut = ContainerResolver(self.bucketer)
        self.user = InternalHackleUser.builder().identifier("$id", "test_id").build()

    def test_식별자가_없는경우_False(self):
        # given
        experiment = create_experiment(identifier_type='custom_id')
        request = create_experiment_request(experiment=experiment, user=self.user)

        # when
        actual = self.sut.is_user_in_container_group(request, Mock())

        # then
        self.assertFalse(actual)

    def test_Bucket이_없으면_에러(self):
        # given
        workspace = Mock(spec=Workspace)
        workspace.get_bucket_or_none.return_value = None
        container = Container(1, 42, [])

        request = create_experiment_request(workspace=workspace, user=self.user)

        # when
        with self.assertRaises(Exception) as actual:
            self.sut.is_user_in_container_group(request, container)

        # then
        self.assertEqual('Bucket[42]', str(actual.exception))

    def test_bucketing_결과_slot에_할당받지_않았으면_False(self):
        # given
        bucket = Mock(spec=Bucket)
        workspace = Mock(spec=Workspace)
        workspace.get_bucket_or_none.return_value = bucket
        container = Container(1, 42, [])

        self.bucketer.bucketing.return_value = None

        request = create_experiment_request(workspace=workspace, user=self.user)

        # when
        actual = self.sut.is_user_in_container_group(request, container)

        # then
        self.assertFalse(actual)

    def test_bucketing_결과에_해당하는_container_group_정보를_못찾는_경우_예외(self):
        # given
        bucket = Mock(spec=Bucket)
        workspace = Mock(spec=Workspace)
        workspace.get_bucket_or_none.return_value = bucket
        container = Container(1, 42, [])

        slot = Slot(0, 100, 320)
        self.bucketer.bucketing.return_value = slot

        request = create_experiment_request(workspace=workspace, user=self.user)

        # when
        with self.assertRaises(Exception) as actual:
            self.sut.is_user_in_container_group(request, container)

        # then
        self.assertEqual('ContainerGroup[320]', str(actual.exception))

    def test_할당된_container_group에_실험이_없으면_False(self):
        # given
        bucket = Mock(spec=Bucket)
        workspace = Mock(spec=Workspace)
        workspace.get_bucket_or_none.return_value = bucket
        container = Container(1, 42, [ContainerGroup(320, [1000])])

        slot = Slot(0, 100, 320)
        self.bucketer.bucketing.return_value = slot

        experiment = create_experiment(id=999)
        request = create_experiment_request(workspace=workspace, experiment=experiment, user=self.user)

        # when
        actual = self.sut.is_user_in_container_group(request, container)

        # then
        self.assertFalse(actual)

    def test_할당된_container_group에_실험이_있으면_True(self):
        # given
        bucket = Mock(spec=Bucket)
        workspace = Mock(spec=Workspace)
        workspace.get_bucket_or_none.return_value = bucket
        container = Container(1, 42, [ContainerGroup(320, [999])])

        slot = Slot(0, 100, 320)
        self.bucketer.bucketing.return_value = slot

        experiment = create_experiment(id=999)
        request = create_experiment_request(workspace=workspace, experiment=experiment, user=self.user)

        # when
        actual = self.sut.is_user_in_container_group(request, container)

        # then
        self.assertTrue(actual)



from hackle.internal.logger.log import Log
from hackle.internal.type import hackle_types


class IdentifiersBuilder(object):
    __MAX_IDENTIFIER_TYPE_LENGTH = 128
    __MAX_IDENTIFIER_VALUE_LENGTH = 512

    def __init__(self):
        self.__identifiers = {}

    def build(self):
        return self.__identifiers

    def add_identifiers(self, identifiers):
        if identifiers is None:
            return self

        if not isinstance(identifiers, dict):
            Log.get().warning('Identifiers MUST be dictionary [Request Type={}]'.format(type(identifiers)))
            return self

        for identifier_type in identifiers:
            self.add(identifier_type, identifiers[identifier_type])

        return self

    def add(self, identifier_type, identifier_value):

        sanitized_identifier_value = self.sanitize_value_or_none(identifier_value)
        if self.__is_value_type(identifier_type) and sanitized_identifier_value is not None:
            self.__identifiers[identifier_type] = sanitized_identifier_value
        else:
            Log.get().warning('Invalid user identifier [type={}, value={}]'.format(identifier_type, identifier_value))
        return self

    @staticmethod
    def sanitize_value_or_none(identifier_value):
        if identifier_value is None:
            return None

        if hackle_types.is_not_empty_string(identifier_value) \
                and len(identifier_value) <= IdentifiersBuilder.__MAX_IDENTIFIER_VALUE_LENGTH:
            return identifier_value

        if hackle_types.is_finite_number(identifier_value):
            return str(identifier_value)

        return None

    def __is_value_type(self, identifier_type):
        if identifier_type is None:
            return False

        if not hackle_types.is_string(identifier_type):
            return False

        if hackle_types.is_empty_string(identifier_type):
            return False

        if len(identifier_type) > self.__MAX_IDENTIFIER_TYPE_LENGTH:
            return False

        return True



from hackle.decision import ExperimentDecision, FeatureFlagDecision, DecisionReason, RemoteConfigDecision
from hackle.internal.evaluation.evaluator.delegating_evaluator import DelegatingEvaluator
from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.evaluation.evaluator.experiment.experiment_evaluator import ExperimentEvaluator
from hackle.internal.evaluation.evaluator.experiment.experiment_request import ExperimentRequest
from hackle.internal.evaluation.evaluator.remoteconfig.remote_config_evaluator import RemoteConfigEvaluator
from hackle.internal.evaluation.evaluator.remoteconfig.remote_config_request import RemoteConfigRequest
from hackle.internal.evaluation.flow.evaluation_flow_factory import EvaluationFlowFactory
from hackle.internal.event.event_processor import EventProcessor
from hackle.internal.event.user_event import UserEvent
from hackle.internal.event.user_event_factory import UserEventFactory
from hackle.internal.model.entities import EventType
from hackle.internal.time.clock import SYSTEM_CLOCK
from hackle.internal.user.internal_hackle_user import InternalHackleUser
from hackle.internal.workspace.workspace_fetcher import WorkspaceFetcher
from hackle.model import HackleEvent


class HackleCore(object):

    def __init__(self,
                 experiment_evaluator,
                 remote_config_evaluator,
                 workspace_fetcher,
                 event_factory,
                 event_processor):
        """
        :param ExperimentEvaluator experiment_evaluator:
        :param RemoteConfigEvaluator remote_config_evaluator:
        :param WorkspaceFetcher workspace_fetcher:
        :param UserEventFactory event_factory:
        :param EventProcessor event_processor:
        """
        self.__experiment_evaluator = experiment_evaluator
        self.__remote_config_evaluator = remote_config_evaluator
        self.__workspace_fetcher = workspace_fetcher
        self.__event_factory = event_factory
        self.__event_processor = event_processor
        self.__clock = SYSTEM_CLOCK

    @staticmethod
    def create(workspace_fetcher, event_processor):
        """
        :param WorkspaceFetcher workspace_fetcher:
        :param EventProcessor event_processor:
        :rtype: HackleCore
        """
        delegating_evaluator = DelegatingEvaluator()
        evaluation_flow_factory = EvaluationFlowFactory(delegating_evaluator)

        experiment_evaluator = ExperimentEvaluator(evaluation_flow_factory)
        remote_config_evaluator = RemoteConfigEvaluator(
            evaluation_flow_factory.remote_config_parameter_target_rule_determiner)

        delegating_evaluator.add(experiment_evaluator)
        delegating_evaluator.add(remote_config_evaluator)

        return HackleCore(
            experiment_evaluator,
            remote_config_evaluator,
            workspace_fetcher,
            UserEventFactory(),
            event_processor
        )

    def close(self):
        self.__workspace_fetcher.stop()
        self.__event_processor.stop()

    def experiment(self, experiment_key, user, default_variation_key):
        """
        :param int experiment_key:
        :param InternalHackleUser user:
        :param str default_variation_key:

        :rtype: ExperimentDecision
        """
        workspace = self.__workspace_fetcher.fetch()
        if workspace is None:
            return ExperimentDecision(default_variation_key, DecisionReason.SDK_NOK_READY)

        experiment = workspace.get_experiment_or_none(experiment_key)
        if experiment is None:
            return ExperimentDecision(default_variation_key, DecisionReason.EXPERIMENT_NOT_FOUND)

        request = ExperimentRequest.of(workspace, user, experiment, default_variation_key)
        evaluation = self.__experiment_evaluator.evaluate(request, Evaluator.context())

        events = self.__event_factory.create(request, evaluation)
        for event in events:
            self.__event_processor.process(event)

        return ExperimentDecision(evaluation.variation_key, evaluation.reason, evaluation.config)

    def feature_flag(self, feature_key, user):
        """
        :param int feature_key:
        :param InternalHackleUser user:

        :rtype: FeatureFlagDecision
        """
        workspace = self.__workspace_fetcher.fetch()
        if workspace is None:
            return FeatureFlagDecision(False, DecisionReason.SDK_NOK_READY)

        feature_flag = workspace.get_feature_flag_or_none(feature_key)
        if feature_flag is None:
            return FeatureFlagDecision(False, DecisionReason.FEATURE_FLAG_NOT_FOUND)

        request = ExperimentRequest.of(workspace, user, feature_flag, 'A')
        evaluation = self.__experiment_evaluator.evaluate(request, Evaluator.context())

        events = self.__event_factory.create(request, evaluation)
        for event in events:
            self.__event_processor.process(event)

        if evaluation.variation_key == 'A':
            return FeatureFlagDecision(False, evaluation.reason, evaluation.config)
        else:
            return FeatureFlagDecision(True, evaluation.reason, evaluation.config)

    def track(self, event, user):
        """
        :param HackleEvent event:
        :param InternalHackleUser user:
        """
        event_type = self.__event_type(event)
        self.__event_processor.process(UserEvent.track(user, event_type, event, self.__clock.current_millis()))
        return

    def __event_type(self, event):
        """
        :param hackle.model.HackleEvent event:

        :rtype: hackle.internal.model.entities.EventType
        """
        workspace = self.__workspace_fetcher.fetch()

        if workspace is None:
            return EventType(0, event.key)

        event_type = workspace.get_event_type_or_none(event.key)

        if event_type is None:
            return EventType(0, event.key)

        return event_type

    def remote_config(self, key, user, required_type, default_value):
        """
        :param str key:
        :param InternalHackleUser user:
        :param str required_type:
        :param object or None default_value:

        :rtype: RemoteConfigDecision
        """
        workspace = self.__workspace_fetcher.fetch()
        if workspace is None:
            return RemoteConfigDecision(default_value, DecisionReason.SDK_NOK_READY)

        parameter = workspace.get_remote_config_parameter_or_none(key)
        if parameter is None:
            return RemoteConfigDecision(default_value, DecisionReason.REMOTE_CONFIG_PARAMETER_NOT_FOUND)

        request = RemoteConfigRequest(workspace, user, parameter, required_type, default_value)
        evaluation = self.__remote_config_evaluator.evaluate(request, Evaluator.context())

        events = self.__event_factory.create(request, evaluation)
        for event in events:
            self.__event_processor.process(event)

        return RemoteConfigDecision(evaluation.value, evaluation.reason)



from .decision import RemoteConfigDecision, DecisionReason
from .internal.hackle_core import HackleCore
from .internal.type import hackle_types
from .internal.user.hackle_user_resolver import HackleUserResolver
from .model import HackleRemoteConfig, HackleUser


class HackleRemoteConfigImpl(HackleRemoteConfig):
    def __init__(self, user, core, hackle_user_resolver):
        """
        :param HackleUser user:
        :param HackleCore core:
        :param HackleUserResolver hackle_user_resolver:
        """
        self.__user = user
        self.__core = core
        self.__hackle_user_resolver = hackle_user_resolver

    def get(self, key, default=None):
        if hackle_types.is_string(default):
            parameter_value = self.__get(key, 'STRING', default).value
        elif hackle_types.is_number(default):
            parameter_value = self.__get(key, 'NUMBER', default).value
        elif hackle_types.is_bool(default):
            parameter_value = self.__get(key, 'BOOLEAN', default).value
        elif default is None:
            parameter_value = self.__get(key, 'NULL', default).value
        else:
            parameter_value = self.__get(key, 'UNKNOWN', default).value

        return parameter_value

    def __get(self, key, required_type, default):
        hackle_user = self.__hackle_user_resolver.resolve_or_none(self.__user)

        if hackle_user is None:
            return RemoteConfigDecision(default, DecisionReason.INVALID_INPUT)

        if key is None:
            return RemoteConfigDecision(default, DecisionReason.INVALID_INPUT)

        return self.__core.remote_config(key, hackle_user, required_type, default)



from unittest import TestCase
from unittest.mock import Mock

from hackle.decision import DecisionReason
from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.evaluation.evaluator.experiment.experiment_evaluation import ExperimentEvaluation
from hackle.internal.evaluation.evaluator.experiment.experiment_request import ExperimentRequest
from hackle.internal.model.entities import ParameterConfiguration
from hackle.internal.user.internal_hackle_user import InternalHackleUser
from hackle.internal.workspace.workspace import Workspace
from tests.internal.model.test_entities import create_experiment, create_variation


class ExperimentEvaluationTest(TestCase):

    def test_create_by_variation(self):
        experiment = create_experiment(
            id=42,
            key=50,
            variations=[
                create_variation(id=320, key='A', config_id=99),
                create_variation(id=321, key='B', config_id=100),
            ]
        )

        variation = experiment.get_variation_by_id_or_none(321)

        config = Mock(sepc=ParameterConfiguration)
        workspace = Mock(spec=Workspace)
        workspace.get_parameter_configuration_or_none.return_value = config

        user = InternalHackleUser.builder().build()
        request = ExperimentRequest.of(workspace, user, experiment, 'H')

        context = Evaluator.context()
        context.add_evaluation(Mock(spec=Evaluator.Evaluation))

        evaluation = ExperimentEvaluation.of(request, context, variation, DecisionReason.TRAFFIC_ALLOCATED)

        self.assertEqual(DecisionReason.TRAFFIC_ALLOCATED, evaluation.reason)
        self.assertEqual(1, len(evaluation.target_evaluations))
        self.assertEqual(experiment, evaluation.experiment)
        self.assertEqual(321, evaluation.variation_id)
        self.assertEqual('B', evaluation.variation_key)
        self.assertEqual(config, evaluation.config)

    def test_create_by_variation_config_none(self):
        experiment = create_experiment(
            id=42,
            key=50,
            variations=[
                create_variation(id=320, key='A'),
                create_variation(id=321, key='B'),
            ]
        )

        variation = experiment.get_variation_by_id_or_none(321)

        workspace = Mock(spec=Workspace)

        user = InternalHackleUser.builder().build()
        request = ExperimentRequest.of(workspace, user, experiment, 'H')

        context = Evaluator.context()
        context.add_evaluation(Mock(spec=Evaluator.Evaluation))

        evaluation = ExperimentEvaluation.of(request, context, variation, DecisionReason.TRAFFIC_ALLOCATED)

        self.assertEqual(DecisionReason.TRAFFIC_ALLOCATED, evaluation.reason)
        self.assertEqual(1, len(evaluation.target_evaluations))
        self.assertEqual(experiment, evaluation.experiment)
        self.assertEqual(321, evaluation.variation_id)
        self.assertEqual('B', evaluation.variation_key)
        self.assertIsNone(evaluation.config)

    def test_create_by_default(self):
        experiment = create_experiment(
            id=42,
            key=50,
            variations=[
                create_variation(id=320, key='A'),
                create_variation(id=321, key='B'),
            ]
        )

        workspace = Mock(spec=Workspace)

        user = InternalHackleUser.builder().build()
        request = ExperimentRequest.of(workspace, user, experiment, 'A')

        evaluation = ExperimentEvaluation.of_default(request, Evaluator.context(), DecisionReason.TRAFFIC_NOT_ALLOCATED)

        self.assertEqual(DecisionReason.TRAFFIC_NOT_ALLOCATED, evaluation.reason)
        self.assertEqual(0, len(evaluation.target_evaluations))
        self.assertEqual(experiment, evaluation.experiment)
        self.assertEqual(320, evaluation.variation_id)
        self.assertEqual('A', evaluation.variation_key)
        self.assertIsNone(evaluation.config)

    def test_create_by_default_null(self):
        experiment = create_experiment(
            id=42,
            key=50,
            variations=[
                create_variation(id=320, key='A'),
                create_variation(id=321, key='B'),
            ]
        )

        workspace = Mock(spec=Workspace)

        user = InternalHackleUser.builder().build()
        request = ExperimentRequest.of(workspace, user, experiment, 'C')

        evaluation = ExperimentEvaluation.of_default(request, Evaluator.context(), DecisionReason.TRAFFIC_NOT_ALLOCATED)

        self.assertEqual(DecisionReason.TRAFFIC_NOT_ALLOCATED, evaluation.reason)
        self.assertEqual(0, len(evaluation.target_evaluations))
        self.assertEqual(experiment, evaluation.experiment)
        self.assertEqual(None, evaluation.variation_id)
        self.assertEqual('C', evaluation.variation_key)
        self.assertIsNone(evaluation.config)

    def test_copy_with(self):
        evaluation = ExperimentEvaluation(DecisionReason.TRAFFIC_ALLOCATED, Mock(), Mock(), 42, 'A', Mock())

        actual = evaluation.copy_with(DecisionReason.TRAFFIC_ALLOCATED_BY_TARGETING)

        self.assertEqual(DecisionReason.TRAFFIC_ALLOCATED_BY_TARGETING, actual.reason)
        self.assertEqual(evaluation.target_evaluations, actual.target_evaluations)
        self.assertEqual(evaluation.experiment, actual.experiment)
        self.assertEqual(evaluation.variation_id, actual.variation_id)
        self.assertEqual(evaluation.variation_key, actual.variation_key)
        self.assertEqual(evaluation.config, actual.config)



from unittest import TestCase
from unittest.mock import Mock

from hackle.internal.evaluation.bucket.bucketer import Bucketer
from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.evaluation.match.target_matcher import TargetMatcher
from hackle.internal.evaluation.target.remote_config_target_rule_determiner import RemoteConfigTargetRuleDeterminer, \
    RemoteConfigTargetRuleMatcher
from hackle.internal.model.entities import *
from hackle.internal.workspace.workspace import Workspace
from tests.internal.evaluation.evaluator.remoteconfig.test_remote_config_request import create_remote_config_request
from tests.internal.model.test_entities import create_parameter


class RemoteConfigTargetRuleDeterminerTest(TestCase):

    def setUp(self):
        self.matches = []
        self.matcher = Mock(spec=RemoteConfigTargetRuleMatcher)
        self.matcher.matches = Mock(side_effect=self.matches)
        self.sut = RemoteConfigTargetRuleDeterminer(self.matcher)

    def test_등록된_rule이_없는경우_None리턴(self):
        # given
        parameter = create_parameter()
        request = create_remote_config_request(parameter=parameter)

        # when
        actual = self.sut.determine_target_rule_or_none(request, Evaluator.context())

        # then
        self.assertIsNone(actual)

    def test_등록된_규칙중_일치하는_첫번쨰_규칙을_리턴한다(self):
        # given
        parameter = create_parameter(target_rules=[
            self.__target_rule(False),
            self.__target_rule(False),
            self.__target_rule(False),
            self.__target_rule(False),
            self.__target_rule(True),
            self.__target_rule(False),
        ])
        request = create_remote_config_request(parameter=parameter)

        # when
        actual = self.sut.determine_target_rule_or_none(request, Evaluator.context())

        # then
        self.assertEqual(parameter.target_rules[4], actual)
        self.assertEqual(5, self.matcher.matches.call_count)

    def test_등록된_규칙중_일치하는_규칙이_하나도_없는경우_None리턴(self):
        parameter = create_parameter(target_rules=[
            self.__target_rule(False),
            self.__target_rule(False),
            self.__target_rule(False),
            self.__target_rule(False),
            self.__target_rule(False),
            self.__target_rule(False),
        ])
        request = create_remote_config_request(parameter=parameter)

        # when
        actual = self.sut.determine_target_rule_or_none(request, Evaluator.context())

        # then
        self.assertIsNone(actual)
        self.assertEqual(6, self.matcher.matches.call_count)

    def __target_rule(self, is_match):
        self.matches.append(is_match)
        target_rule = Mock(spec=RemoteConfigTargetRule)
        return target_rule


class RemoteConfigTargetRuleMatcherTest(TestCase):

    def setUp(self):
        self.target_matcher = Mock(spec=TargetMatcher)
        self.bucketer = Mock(spec=Bucketer)
        self.sut = RemoteConfigTargetRuleMatcher(self.target_matcher, self.bucketer)

    def test_target_rule_not_matched(self):
        target_rule = self.__target_rule(False)
        request = create_remote_config_request()

        actual = self.sut.matches(request, Evaluator.context(), target_rule)

        self.assertFalse(actual)

    def test_identifier_not_found(self):
        target_rule = self.__target_rule(True)
        request = create_remote_config_request(parameter=create_parameter(identifier_type='custom_id'))

        actual = self.sut.matches(request, Evaluator.context(), target_rule)

        self.assertFalse(actual)

    def test_bucket_not_found(self):
        target_rule = self.__target_rule(True)
        target_rule.bucket_id = 42

        request = create_remote_config_request()

        with self.assertRaises(Exception) as actual:
            self.sut.matches(request, Evaluator.context(), target_rule)

        self.assertEqual('Bucket[42]', str(actual.exception))

    def test_not_allocated(self):
        target_rule = self.__target_rule(True)
        target_rule.bucket_id = 42

        workspace = Mock(spec=Workspace)
        workspace.get_bucket_or_none.return_value = Mock()

        self.bucketer.bucketing.return_value = None

        request = create_remote_config_request(workspace=workspace, parameter=create_parameter())

        actual = self.sut.matches(request, Evaluator.context(), target_rule)

        self.assertFalse(actual)

    def test_allocated(self):
        target_rule = self.__target_rule(True)
        target_rule.bucket_id = 42

        workspace = Mock(spec=Workspace)
        workspace.get_bucket_or_none.return_value = Mock()

        self.bucketer.bucketing.return_value = Mock()

        request = create_remote_config_request(workspace=workspace, parameter=create_parameter())

        actual = self.sut.matches(request, Evaluator.context(), target_rule)

        self.assertTrue(actual)

    def __target_rule(self, is_match):
        self.target_matcher.matches.return_value = is_match
        target_rule = Mock(spec=RemoteConfigTargetRule)
        target_rule.target = Mock()
        return target_rule



from unittest import TestCase
from unittest.mock import Mock

from hackle.decision import DecisionReason
from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.evaluation.evaluator.experiment.experiment_evaluation import ExperimentEvaluation
from hackle.internal.evaluation.evaluator.remoteconfig.remote_config_evaluation import RemoteConfigEvaluation
from hackle.internal.event.event_dispatcher import EventDispatcher
from hackle.internal.event.user_event import UserEvent
from hackle.internal.http.http_client import HttpClient
from hackle.internal.model.entities import EventType
from hackle.internal.model.properties_builder import PropertiesBuilder
from hackle.internal.user.identifier_type import IdentifierType
from hackle.internal.user.internal_hackle_user import InternalHackleUser
from hackle.model import HackleEvent
from tests.internal.evaluation.evaluator.experiment.test_experiment_request import create_experiment_request
from tests.internal.evaluation.evaluator.remoteconfig.test_remote_config_request import create_remote_config_request


class EventDispatcherTest(TestCase):

    def test_dispatch(self):
        user = InternalHackleUser.builder() \
            .identifier(IdentifierType.ID, "id") \
            .identifier(IdentifierType.USER, "user_id") \
            .identifier(IdentifierType.DEVICE, "device_id") \
            .property('age', 42) \
            .build()

        experiment_evaluation = ExperimentEvaluation.of_default(
            create_experiment_request(), Evaluator.context(), DecisionReason.TRAFFIC_ALLOCATED)
        exposure_event = UserEvent.exposure(user, experiment_evaluation, {}, 42)

        event = HackleEvent.builder('test').property('amount', 4200).build()
        track_event = UserEvent.track(user, EventType(2, 'test'), event, 43)

        remote_config_evaluation = RemoteConfigEvaluation.of_default(
            create_remote_config_request(), Evaluator.context(), DecisionReason.DEFAULT_RULE, PropertiesBuilder())
        remote_config_event = UserEvent.remote_config(user, remote_config_evaluation, {}, 44)

        http_client = Mock(spec=HttpClient)
        sut = EventDispatcher('localhost', http_client)

        sut.dispatch([exposure_event, track_event, remote_config_event])

        http_client.post.assert_called_once()
        self.assertEqual('localhost/api/v2/events', http_client.post.call_args[0][0])
        a = len(http_client.post.call_args[0][1])
        self.assertGreater(len(http_client.post.call_args[0][1]), 1000)



from unittest import TestCase
from unittest.mock import Mock

from hackle.decision import DecisionReason
from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.evaluation.evaluator.experiment.experiment_evaluation import ExperimentEvaluation
from hackle.internal.evaluation.match.experiment_condition_matcher import FeatureFlagConditionMatcher
from hackle.internal.evaluation.match.value_operator_matcher import ValueOperatorMatcher
from hackle.internal.model.entities import TargetCondition, TargetKey, TargetMatch
from hackle.internal.workspace.workspace import Workspace
from tests.internal.evaluation.evaluator.experiment.test_experiment_request import create_experiment_request
from tests.internal.model.test_entities import create_experiment


class FeatureFlagConditionMatcherTest(TestCase):

    def setUp(self):
        self.evaluator = Mock(spec=Evaluator)
        self.value_operator_matcher = Mock(spec=ValueOperatorMatcher)
        self.sut = FeatureFlagConditionMatcher(self.evaluator, self.value_operator_matcher)
        self.context = Evaluator.context()

    def test_key_is_not_int(self):
        # given
        experiment = create_experiment(type='FEATURE_FLAG')
        request = create_experiment_request(experiment=experiment)
        condition = TargetCondition(
            TargetKey('FEATURE_FLAG', 'string'),
            TargetMatch('MATCH', 'IN', 'BOOLEAN', [True])
        )

        # when
        with self.assertRaises(Exception) as actual:
            self.sut.matches(request, Evaluator.context(), condition)

        # then
        self.assertEqual('Invalid key [FEATURE_FLAG, string]', str(actual.exception))

    def test_experiment_not_found(self):
        # given
        experiment = create_experiment(type='FEATURE_FLAG')
        request = create_experiment_request(experiment=experiment)
        condition = TargetCondition(
            TargetKey('FEATURE_FLAG', '42'),
            TargetMatch('MATCH', 'IN', 'BOOLEAN', [True])
        )

        # when
        actual = self.sut.matches(request, Evaluator.context(), condition)

        # then
        self.assertFalse(actual)

    # noinspection PyMethodMayBeStatic
    def __request(self, experiment):
        workspace = Mock(spec=Workspace)
        workspace.get_feature_flag_or_none.return_value = experiment
        return create_experiment_request(workspace=workspace, experiment=experiment)

    def test_new_evaluation(self):
        # given
        experiment = create_experiment(type='FEATURE_FLAG')
        request = self.__request(experiment)
        condition = TargetCondition(
            TargetKey('FEATURE_FLAG', '42'),
            TargetMatch('MATCH', 'IN', 'BOOLEAN', [True])
        )

        evaluation = ExperimentEvaluation.of(request, self.context, request.experiment.variations[0],
                                             DecisionReason.DEFAULT_RULE)
        self.evaluator.evaluate.return_value = evaluation
        self.value_operator_matcher.matches.return_value = True

        # when
        actual = self.sut.matches(request, self.context, condition)

        # then
        self.assertEqual(True, actual)
        self.evaluator.evaluate.assert_called_once()
        self.assertEqual(evaluation, self.context.get_evaluation_or_none(experiment))

    def test_alreadt_evaluated(self):
        # given
        experiment = create_experiment(type='FEATURE_FLAG')
        request = self.__request(experiment)
        condition = TargetCondition(
            TargetKey('FEATURE_FLAG', '42'),
            TargetMatch('MATCH', 'IN', 'BOOLEAN', [True])
        )

        evaluation = ExperimentEvaluation.of(request, self.context, request.experiment.variations[0],
                                             DecisionReason.DEFAULT_RULE)
        self.context.add_evaluation(evaluation)
        self.value_operator_matcher.matches.return_value = True

        # when
        actual = self.sut.matches(request, self.context, condition)

        # then
        self.assertEqual(True, actual)
        self.evaluator.evaluate.assert_not_called()
        self.assertEqual(evaluation, self.context.get_evaluation_or_none(experiment))



from unittest import TestCase
from unittest.mock import Mock

from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.evaluation.evaluator.experiment.experiment_evaluation import ExperimentEvaluation
from hackle.internal.evaluation.evaluator.experiment.experiment_evaluator import ExperimentEvaluator
from hackle.internal.evaluation.evaluator.experiment.experiment_request import ExperimentRequest
from hackle.internal.evaluation.evaluator.remoteconfig.remote_config_request import RemoteConfigRequest
from hackle.internal.evaluation.flow.evaluation_flow import EvaluationFlow
from hackle.internal.evaluation.flow.evaluation_flow_factory import EvaluationFlowFactory
from tests.internal.evaluation.evaluator.experiment.test_experiment_request import create_experiment_request


class ExperimentEvaluatorTest(TestCase):

    def setUp(self):
        self.evaluation_flow_factory = Mock(spec=EvaluationFlowFactory)
        self.sut = ExperimentEvaluator(self.evaluation_flow_factory)

    def test_supports(self):
        self.assertTrue(self.sut.supports(Mock(spec=ExperimentRequest)))
        self.assertFalse(self.sut.supports(Mock(spec=RemoteConfigRequest)))

    def test__evaluate__circular_call(self):
        request = create_experiment_request()
        context = Evaluator.context()
        context.add_request(request)

        with self.assertRaises(Exception) as actual:
            self.sut.evaluate(request, context)

        self.assertIn('Circular evaluation has occurred', str(actual.exception))

    def test__evaluate__flow(self):
        evaluation = Mock(spec=ExperimentEvaluation)
        evaluation_flow = Mock(spec=EvaluationFlow)
        evaluation_flow.evaluate.return_value = evaluation

        self.evaluation_flow_factory.get_evaluation_flow.return_value = evaluation_flow

        request = create_experiment_request()
        context = Evaluator.context()

        actual = self.sut.evaluate(request, context)

        self.assertEqual(evaluation, actual)
        self.assertEqual(0, len(context.stack))



from hackle.internal.logger.log import Log
from hackle.internal.type import hackle_types


class PropertiesBuilder(object):
    __SYSTEM_PROPERTY_KEY_PREFIX = '$'
    __MAX_PROPERTIES_COUNT = 128
    __MAX_PROPERTY_KEY_LENGTH = 128
    __MAX_PROPERTY_VALUE_LENGTH = 1024

    def __init__(self):
        self.__properties = {}  # type: dict[str, object]

    @staticmethod
    def sanitize(properties):
        """
        :param dict or None properties:
        :rtype: dict
        """
        return PropertiesBuilder().add_properties(properties or {}).build()

    def build(self):
        """
        :rtype: dict
        """
        return self.__properties

    def add_properties(self, properties):
        """
        :param dict properties:
        :rtype: PropertiesBuilder
        """
        if properties is None or not isinstance(properties, dict):
            Log.get().warning('Invalid properties: {} (expected: dict)'.format(properties))
            return self

        for property_key in properties:
            self.add(property_key, properties[property_key])

        return self

    def add(self, key, value):
        """
        :param str key:
        :param object value:
        :rtype: PropertiesBuilder
        """
        if len(self.__properties) >= self.__MAX_PROPERTIES_COUNT:
            return self

        if not self.__is_valid_key(key):
            Log.get().warning('Invalid property key: {} (expected: string[1..128])')
            return self

        sanitized_value = self.__sanitize_value_or_none(key, value)
        if sanitized_value is None:
            return self

        self.__properties[key] = sanitized_value

        return self

    def __is_valid_key(self, key):
        if key is None:
            return False

        if not hackle_types.is_string(key):
            return False

        if len(key) == 0:
            return False

        if len(key) > self.__MAX_PROPERTY_KEY_LENGTH:
            return False

        return True

    def __sanitize_value_or_none(self, key, value):
        """
        :param str key:
        :param object or None value:
        :rtype: object or None
        """
        if value is None:
            return None

        if isinstance(value, list):
            values = []
            for element in value:
                if element is not None and self.__is_valid_element(element):
                    values.append(element)
            return values

        if self.__is_valid_value(value):
            return value

        if key.startswith(self.__SYSTEM_PROPERTY_KEY_PREFIX):
            return value

        return None

    def __is_valid_element(self, element):
        if hackle_types.is_string(element):
            return len(element) <= self.__MAX_PROPERTY_VALUE_LENGTH
        elif hackle_types.is_finite_number(element):
            return True
        else:
            return False

    def __is_valid_value(self, value):
        if hackle_types.is_string(value):
            return len(value) <= self.__MAX_PROPERTY_VALUE_LENGTH
        elif hackle_types.is_finite_number(value):
            return True
        elif hackle_types.is_bool(value):
            return True
        else:
            return False

    def __contains__(self, item):
        return item in self.__properties



from unittest import TestCase
from unittest.mock import Mock

from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.evaluation.evaluator.experiment.experiment_evaluation import ExperimentEvaluation
from hackle.internal.evaluation.evaluator.remoteconfig.remote_config_evaluation import RemoteConfigEvaluation
from tests.internal.evaluation.evaluator.experiment.test_experiment_request import create_experiment_request
from tests.internal.model.test_entities import create_experiment


class EvaluatorContextTest(TestCase):
    def test_stack(self):
        context = Evaluator.context()
        self.assertEqual(0, len(context.stack))

        request_1 = create_experiment_request(experiment=create_experiment(id=1))
        context.add_request(request_1)
        stack_1 = context.stack
        self.assertEqual(1, len(stack_1))

        request_2 = create_experiment_request(experiment=create_experiment(id=2))
        context.add_request(request_2)
        stack_2 = context.stack
        self.assertEqual(2, len(stack_2))

        context.remove_request(request_2)
        self.assertEqual(1, len(context.stack))

        context.remove_request(request_1)
        self.assertEqual(0, len(context.stack))

        self.assertEqual(1, len(stack_1))
        self.assertEqual(2, len(stack_2))

    def test_target_evaluations(self):
        context = Evaluator.context()
        self.assertEqual(0, len(context.target_evaluations))

        experiment = create_experiment(id=1)

        evaluation_1 = Mock(spec=RemoteConfigEvaluation)
        context.add_evaluation(evaluation_1)
        target_evaluations_1 = context.target_evaluations
        self.assertEqual(1, len(target_evaluations_1))
        self.assertEqual(None, context.get_evaluation_or_none(experiment))

        evaluation_2 = Mock(spec=ExperimentEvaluation)
        evaluation_2.experiment = experiment
        context.add_evaluation(evaluation_2)
        target_evaluations_2 = context.target_evaluations
        self.assertEqual(1, len(target_evaluations_1))
        self.assertEqual(2, len(target_evaluations_2))
        self.assertEqual(evaluation_2, context.get_evaluation_or_none(experiment))
        self.assertEqual(None, context.get_evaluation_or_none(create_experiment(id=2)))



from unittest import TestCase

from hackle.decision import DecisionReason
from hackle.internal.evaluation.evaluator.experiment.experiment_evaluation import ExperimentEvaluation
from hackle.internal.evaluation.evaluator.remoteconfig.remote_config_evaluation import RemoteConfigEvaluation
from hackle.internal.event.user_event import UserEvent, ExposureEvent, TrackEvent, RemoteConfigEvent
from hackle.internal.model.entities import ParameterConfiguration, EventType
from hackle.internal.user.internal_hackle_user import InternalHackleUser
from hackle.model import HackleEvent
from tests.internal.model.test_entities import create_experiment, create_parameter


class UserEventTest(TestCase):

    def test_exposure(self):
        config = ParameterConfiguration(42, {})

        experiment = create_experiment()
        evaluation = ExperimentEvaluation(
            DecisionReason.TRAFFIC_ALLOCATED,
            [],
            experiment,
            42,
            'B',
            config
        )
        user = InternalHackleUser.builder().identifier("$id", "test_id").build()

        event = UserEvent.exposure(user, evaluation, {"a": "1"}, 320)

        self.assertIsInstance(event, ExposureEvent)
        self.assertEqual(320, event.timestamp)
        self.assertEqual(user, event.user)
        self.assertEqual(experiment, event.experiment)
        self.assertEqual(42, event.variation_id)
        self.assertEqual('B', event.variation_key)
        self.assertEqual(DecisionReason.TRAFFIC_ALLOCATED, event.reason)
        self.assertEqual({"a": "1"}, event.properties)

    def test_track(self):
        user = InternalHackleUser.builder().identifier("$id", "test_id").build()
        event_type = EventType(320, 'test_event_key')
        event = HackleEvent.builder('test_event_key').build()

        user_event = UserEvent.track(user, event_type, event, 42)

        self.assertIsInstance(user_event, TrackEvent)
        self.assertEqual(42, user_event.timestamp)
        self.assertEqual(user, user_event.user)
        self.assertEqual(event_type, user_event.event_type)
        self.assertEqual(event, user_event.event)

    def test_remote_config(self):
        user = InternalHackleUser.builder().identifier("$id", "test_id").build()
        parameter = create_parameter()

        evaluation = RemoteConfigEvaluation(
            DecisionReason.DEFAULT_RULE,
            [],
            parameter,
            42,
            'final_value',
            {"a": "1"}
        )

        event = UserEvent.remote_config(user, evaluation, {"b": "2"}, 320)

        self.assertIsInstance(event, RemoteConfigEvent)
        self.assertEqual(320, event.timestamp)
        self.assertEqual(user, event.user)
        self.assertEqual(parameter, event.parameter)
        self.assertEqual(42, event.value_id)
        self.assertEqual(DecisionReason.DEFAULT_RULE, event.reason)
        self.assertEqual({"b": "2"}, event.properties)



import abc

from six import add_metaclass, string_types

from hackle.internal.evaluation.match.semantic_version import SemanticVersion
from hackle.internal.type import hackle_types


@add_metaclass(abc.ABCMeta)
class ValueMatcher(object):

    @abc.abstractmethod
    def matches(self, operator_matcher, user_value, match_value):
        pass


class StringValueMatcher(ValueMatcher):
    def matches(self, operator_matcher, user_value, match_value):
        str_user_value = self.__string_or_none(user_value)
        str_match_value = self.__string_or_none(match_value)
        if str_user_value is not None and str_match_value is not None:
            return operator_matcher.string_matches(str_user_value, str_match_value)
        else:
            return False

    @staticmethod
    def __string_or_none(value):
        if isinstance(value, string_types):
            return value
        elif hackle_types.is_finite_number(value):
            return str(value)
        else:
            return None


class NumberValueMatcher(ValueMatcher):
    def matches(self, operator_matcher, user_value, match_value):
        number_user_value = self.__number_or_none(user_value)
        number_match_value = self.__number_or_none(match_value)
        if number_user_value is not None and number_match_value is not None:
            return operator_matcher.number_matches(number_user_value, number_match_value)
        else:
            return False

    @staticmethod
    def __number_or_none(value):
        if hackle_types.is_finite_number(value):
            return value
        elif isinstance(value, string_types):
            try:
                number = float(value)
                if hackle_types.is_finite_number(number):
                    return number
                else:
                    return None
            except ValueError:
                return None
        else:
            return None


class BoolValueMatcher(ValueMatcher):
    def matches(self, operator_matcher, user_value, match_value):
        if isinstance(user_value, bool) and isinstance(match_value, bool):
            return operator_matcher.bool_matches(user_value, match_value)
        else:
            return False


class VersionValueMatcher(ValueMatcher):
    def matches(self, operator_matcher, user_value, match_value):
        user_version = SemanticVersion.parse_or_none(user_value)
        match_version = SemanticVersion.parse_or_none(match_value)
        if user_version is not None and match_version is not None:
            return operator_matcher.version_matches(user_version, match_version)
        else:
            return False


class NoneValueMatcher(ValueMatcher):
    def matches(self, operator_matcher, user_value, match_value):
        return False



from hackle.internal.evaluation.evaluator.remoteconfig.remote_config_request import RemoteConfigRequest
from hackle.internal.model.entities import RemoteConfigParameter, RemoteConfigParameterValue
from hackle.internal.user.identifier_type import IdentifierType
from hackle.internal.user.internal_hackle_user import InternalHackleUser
from hackle.internal.workspace.workspace import Workspace


def create_remote_config_request(**params):
    return RemoteConfigRequest(
        workspace=params.get('workspace', Workspace('{}')),
        user=params.get('user', InternalHackleUser.builder().identifier(IdentifierType.ID, 'user').build()),
        parameter=params.get('parameter', RemoteConfigParameter(1, 'key', 'STRING', IdentifierType.ID, [],
                                                                RemoteConfigParameterValue(1, 'defaultValue'))),
        required_type=params.get('required_type', 'STRING'),
        default_value=params.get('default_value', 'default value')
    )



from unittest import TestCase
from unittest.mock import Mock

from hackle.internal.evaluation.match.segment_condition_matcher import SegmentMatcher
from hackle.internal.evaluation.match.user_condition_matcher import UserConditionMatcher
from hackle.internal.model.entities import *


class SegmentMatcherTest(TestCase):

    def setUp(self):
        self.condition_matches = []
        self.user_condition_matcher = Mock(spec=UserConditionMatcher)
        self.user_condition_matcher.matches = Mock(side_effect=self.condition_matches)
        self.sut = SegmentMatcher(self.user_condition_matcher)

    def test_Target이_비어있으면_match_false(self):
        # given
        segment = Segment(1, "seg_01", 'USER_PROPERTY', [])

        # when
        actual = self.sut.matches(Mock(), Mock(), segment)

        # then
        self.assertFalse(actual)

    def test_하나라도_매칭되는_Target이_있으면_true(self):
        # given
        segment = self._segment(
            [True, True, True, False],  # False
            [False],  # False
            [True, True]  # True
        )

        # when
        actual = self.sut.matches(Mock(), Mock(), segment)

        # then
        self.assertTrue(actual)

    def test_하나라도_매칭되는_Target이_없으면_true(self):
        # given
        segment = self._segment(
            [True, True, True, False],  # False
            [False],  # False
            [False, True]  # False
        )

        # when
        actual = self.sut.matches(Mock(), Mock(), segment)

        # then
        self.assertFalse(actual)

    def _segment(self, *target_conditions):
        targets = []
        for target_matches in target_conditions:
            conditions = []
            for conditionMatch in target_matches:
                condition = Mock(spec=TargetCondition)
                conditions.append(condition)
                self.condition_matches.append(conditionMatch)
            targets.append(Target(conditions))

        return Segment(42, "seg_01", "USER_PROPERTY", targets)



import logging


class Default(object):
    SDK_URL = 'https://sdk.hackle.io'
    EVENT_URL = 'https://event.hackle.io'


class LogLevels(object):
    NOTSET = logging.NOTSET
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL



import time
from unittest import TestCase

from hackle.internal.concurrent.atomic.atomic_integer import AtomicInteger
from hackle.internal.concurrent.schedule.thread_scheduler import ThreadScheduler
from hackle.internal.time.time_unit import MILLISECONDS


class ThreadSchedulerTest(TestCase):
    def test_schedule(self):
        count = AtomicInteger()

        def task():
            count.add_and_get(1)

        scheduler = ThreadScheduler()
        job = scheduler.schedule_periodically(50, 100, MILLISECONDS, task)
        time.sleep(1)
        job.cancel()
        self.assertEqual(10, count.get())

    def test_ignore_exception(self):
        count = AtomicInteger()

        def task():
            count.add_and_get(1)
            raise Exception()

        scheduler = ThreadScheduler()
        job = scheduler.schedule_periodically(50, 100, MILLISECONDS, task)
        time.sleep(1)
        job.cancel()
        self.assertEqual(10, count.get())

    def test_cancel_during_delay(self):
        count = AtomicInteger()

        def task():
            count.add_and_get(1)

        scheduler = ThreadScheduler()
        job = scheduler.schedule_periodically(1000, 100, MILLISECONDS, task)
        time.sleep(0.5)
        job.cancel()
        self.assertEqual(0, count.get())

    def test_long_task(self):
        count = AtomicInteger()

        def long_task():
            time.sleep(0.4)
            count.add_and_get(1)

        scheduler = ThreadScheduler()
        job = scheduler.schedule_periodically(0, 100, MILLISECONDS, long_task)
        time.sleep(1)
        job.cancel()
        self.assertEqual(2, count.get())



from hackle.internal.evaluation.evaluator.evaluator import ContextualEvaluator, Evaluator
from hackle.internal.evaluation.evaluator.experiment.experiment_evaluation import ExperimentEvaluation
from hackle.internal.evaluation.evaluator.experiment.experiment_request import ExperimentRequest
from hackle.internal.evaluation.flow.evaluation_flow_factory import EvaluationFlowFactory


class ExperimentEvaluator(ContextualEvaluator):

    def __init__(self, evaluation_flow_factory):
        """
        :param EvaluationFlowFactory evaluation_flow_factory:
        """
        super(ExperimentEvaluator, self).__init__()
        self.__evaluation_flow_factory = evaluation_flow_factory

    def supports(self, request):
        return isinstance(request, ExperimentRequest)

    def evaluate_internal(self, request, context):
        """
        :param ExperimentRequest request:
        :param Evaluator.Context context:

        :rtype: ExperimentEvaluation
        """
        evaluation_flow = self.__evaluation_flow_factory.get_evaluation_flow(request.experiment.type)
        return evaluation_flow.evaluate(request, context)



import abc

from six import add_metaclass

from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.model.entities import TargetCondition


@add_metaclass(abc.ABCMeta)
class ConditionMatcher(object):

    @abc.abstractmethod
    def matches(self, request, context, condition):
        """
        :param Evaluator.Request request:
        :param Evaluator.Context context:
        :param TargetCondition condition:

        :rtype: bool
        """
        pass



import threading

from six.moves.queue import Queue, Full

from hackle.internal.event.event_dispatcher import EventDispatcher
from hackle.internal.event.user_event import UserEvent
from hackle.internal.logger.log import Log
from hackle.internal.time.time_unit import SECONDS


class EventProcessor(object):
    __SHUTDOWN_SIGNAL = object()
    __FLUSH_SIGNAL = object()

    def __init__(
            self,
            queue,
            event_dispatcher,
            event_dispatch_size,
            flush_scheduler,
            flush_interval_seconds,
            shutdown_timeout_seconds):
        """
        :param Queue queue:
        :param EventDispatcher event_dispatcher:
        :param Scheduler flush_scheduler:
        :param int event_dispatch_size:
        :param float flush_interval_seconds:
        :param float shutdown_timeout_seconds:
        """
        self.__lock = threading.Lock()

        self.__queue = queue
        self.__event_dispatcher = event_dispatcher
        self.__event_dispatch_size = event_dispatch_size
        self.__flush_scheduler = flush_scheduler
        self.__flush_interval_seconds = flush_interval_seconds
        self.__shutdown_timeout_seconds = shutdown_timeout_seconds

        self.__flushing_job = None
        self.__consuming_task = None
        self.__is_started = False

        self.__current_batch = list()

    def process(self, event):
        """
        :param UserEvent event:
        """
        if not isinstance(event, UserEvent):
            Log.get().warning('Invalid event type: {}. (expected: UserEvent)'.format(type(event).__name__))
            return

        try:
            self.__queue.put_nowait(event)
        except Full:
            Log.get().warning('Event not processed. Exceeded event queue capacity. (Current queue size: {})'.format(
                str(self.__queue.qsize())))

    def __flush(self):
        self.__queue.put(self.__FLUSH_SIGNAL)

    def start(self):
        if self.__is_started:
            Log.get().info("EventProcessor is already started")
            return

        self.__consuming_task = threading.Thread(target=self.__consuming)
        self.__consuming_task.daemon = True
        self.__consuming_task.start()

        self.__flushing_job = self.__flush_scheduler.schedule_periodically(
            self.__flush_interval_seconds,
            self.__flush_interval_seconds,
            SECONDS,
            self.__flush
        )

        self.__is_started = True
        Log.get().info('EventProcessor started. Flush events every {} s.'.format(self.__flush_interval_seconds))

    def stop(self):
        if not self.__is_started:
            return

        Log.get().info('Shutting down EventProcessor')

        self.__flushing_job.cancel()
        self.__flush_scheduler.close()

        self.__queue.put(self.__SHUTDOWN_SIGNAL)

        if self.__consuming_task:
            self.__consuming_task.join(self.__shutdown_timeout_seconds)

    def __consuming(self):
        try:
            while True:
                try:
                    message = self.__queue.get(True)

                    if isinstance(message, UserEvent):
                        self.__consume_event(message)
                        continue

                    if message == self.__FLUSH_SIGNAL:
                        self.__dispatch_events()
                        continue

                    if message == self.__SHUTDOWN_SIGNAL:
                        break

                except Exception as e:
                    Log.get().debug('Unexpected exception in event processor: {}'.format(str(e)))

        except Exception as e:
            Log.get().error('Unexpected exception in event processor: {}'.format(str(e)))
        finally:
            Log.get().info('Exit Event Processing loop')
            self.__dispatch_events()

    def __consume_event(self, event):
        with self.__lock:
            self.__current_batch.append(event)

        if len(self.__current_batch) >= self.__event_dispatch_size:
            self.__dispatch_events()

    def __dispatch_events(self):
        batch_len = len(self.__current_batch)
        if batch_len == 0:
            return

        with self.__lock:
            events = list(self.__current_batch)
            self.__current_batch = list()

        try:
            self.__event_dispatcher.dispatch(events)
        except Exception as e:
            Log.get().error('Failed to dispatch events: {}'.format(str(e)))



from hackle.internal.evaluation.match.condition_matcher import ConditionMatcher
from hackle.internal.evaluation.match.value_operator_matcher import ValueOperatorMatcher
from hackle.internal.logger.log import Log
from hackle.internal.model.entities import TargetKey
from hackle.internal.user.internal_hackle_user import InternalHackleUser


class UserConditionMatcher(ConditionMatcher):

    def __init__(self, user_value_resolver, value_operator_matcher):
        """
        :param UserValueResolver user_value_resolver:
        :param ValueOperatorMatcher value_operator_matcher:
        """
        self.__user_value_resolver = user_value_resolver
        self.__value_operator_matcher = value_operator_matcher

    def matches(self, request, context, condition):
        user_value = self.__user_value_resolver.resolve_or_none(request.user, condition.key)
        if user_value is None:
            return False
        return self.__value_operator_matcher.matches(user_value, condition.match)


class UserValueResolver(object):

    # noinspection PyMethodMayBeStatic
    def resolve_or_none(self, user, target_key):
        """
        :param InternalHackleUser user:
        :param TargetKey target_key:

        :rtype: object or None
        """
        if target_key.type == 'USER_ID':
            return user.identifiers.get(target_key.name)
        elif target_key.type == 'USER_PROPERTY':
            return user.properties.get(target_key.name)
        elif target_key.type == 'HACKLE_PROPERTY':
            return None
        elif target_key.type == 'SEGMENT':
            raise Exception('Unsupported target_key.type [SEGMENT]')
        else:
            Log.get().debug('Unsupported type [{}]. Please use the latest version of sdk.'.format(target_key))
            return None



from unittest import TestCase
from unittest.mock import Mock

from hackle.internal.evaluation.action.action_resolver import ActionResolver
from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.evaluation.match.target_matcher import TargetMatcher
from hackle.internal.evaluation.target.override_resolver import OverrideResolver
from hackle.internal.model.entities import Variation, TargetAction, Experiment, TargetRule, Target, TargetCondition, \
    TargetKey, \
    TargetMatch
from hackle.internal.user.internal_hackle_user import InternalHackleUser
from tests.internal.evaluation.evaluator.experiment.test_experiment_request import create_experiment_request


class OverrideResolverTest(TestCase):

    def setUp(self):
        self.target_matcher = Mock(spec=TargetMatcher)
        self.action_resolver = Mock(spec=ActionResolver)
        self.sut = OverrideResolver(self.target_matcher, self.action_resolver)

    def test_identnfer_type에_해당하는_식별자가_없으면_segment_override를_평가한다(self):
        # given
        user = InternalHackleUser(identifiers={'$id': 'test'}, properties={})
        experiment = Experiment(
            id=1,
            key=1,
            type='AB_TEST',
            identifier_type='custom_id',
            status='RUNNING',
            version=1,
            variations=[
                Variation(11, 'A', False),
                Variation(22, 'B', False)
            ],
            user_overrides={'test_user_id': 22},
            segment_overrides=[
                TargetRule(
                    Target([TargetCondition(TargetKey('SEGMENT', "SEGMENT"),
                                            TargetMatch('MATCH', 'IN', 'STRING', ["seg_01"]))]),
                    TargetAction('BUCKET', None, 1)
                )
            ],
            target_audiences=[],
            target_rules=[],
            default_rule=TargetAction('BUCKET', bucket_id=1),
            container_id=None,
            winner_variation_id=None
        )

        self.target_matcher.matches.return_value = True
        self.action_resolver.resolve_or_none.return_value = Variation(11, 'A', False)

        request = create_experiment_request(experiment=experiment, user=user)

        # when
        actual = self.sut.resolve_or_none(request, Evaluator.context())

        # then
        self.assertEqual(Variation(11, 'A', False), actual)
        self.target_matcher.matches.assert_called_once()
        self.action_resolver.resolve_or_none.assert_called_once()

    def test_user_override_되어있지않으면_segment_override를_평가한다(self):
        # given
        user = InternalHackleUser(identifiers={'$id': 'test'}, properties={})
        experiment = Experiment(
            id=1,
            key=1,
            type='AB_TEST',
            identifier_type='$id',
            status='RUNNING',
            version=1,
            variations=[
                Variation(11, 'A', False),
                Variation(22, 'B', False)
            ],
            user_overrides={},
            segment_overrides=[
                TargetRule(
                    Target([TargetCondition(TargetKey('SEGMENT', "SEGMENT"),
                                            TargetMatch('MATCH', 'IN', 'STRING', ["seg_01"]))]),
                    TargetAction('BUCKET', None, 1)
                )
            ],
            target_audiences=[],
            target_rules=[],
            default_rule=TargetAction('BUCKET', bucket_id=1),
            container_id=None,
            winner_variation_id=None
        )

        self.target_matcher.matches.return_value = True
        self.action_resolver.resolve_or_none.return_value = Variation(11, 'A', False)

        request = create_experiment_request(experiment=experiment, user=user)

        # when
        actual = self.sut.resolve_or_none(request, Evaluator.context())

        # then
        self.assertEqual(Variation(11, 'A', False), actual)
        self.target_matcher.matches.assert_called_once()
        self.action_resolver.resolve_or_none.assert_called_once()

    def test_user_override되어있는_경우(self):
        # given
        user = InternalHackleUser(identifiers={'$id': 'test'}, properties={})
        experiment = Experiment(
            id=1,
            key=1,
            type='AB_TEST',
            identifier_type='$id',
            status='RUNNING',
            version=1,
            variations=[
                Variation(11, 'A', False),
                Variation(22, 'B', False)
            ],
            user_overrides={'test': 22},
            segment_overrides=[
                TargetRule(
                    Target([TargetCondition(TargetKey('SEGMENT', "SEGMENT"),
                                            TargetMatch('MATCH', 'IN', 'STRING', ["seg_01"]))]),
                    TargetAction('BUCKET', None, 1)
                )
            ],
            target_audiences=[],
            target_rules=[],
            default_rule=TargetAction('BUCKET', bucket_id=1),
            container_id=None,
            winner_variation_id=None
        )

        request = create_experiment_request(experiment=experiment, user=user)

        # when
        actual = self.sut.resolve_or_none(request, Evaluator.context())

        # then
        self.assertEqual(Variation(22, 'B', False), actual)
        self.target_matcher.matches.assert_not_called()
        self.action_resolver.resolve_or_none.assert_not_called()

    def test_직접입력_Segment_둘다_override_되어있지_않으면_None(self):
        # given
        user = InternalHackleUser(identifiers={'$id': 'test'}, properties={})
        experiment = Experiment(
            id=1,
            key=1,
            type='AB_TEST',
            identifier_type='$id',
            status='RUNNING',
            version=1,
            variations=[
                Variation(11, 'A', False),
                Variation(22, 'B', False)
            ],
            user_overrides={},
            segment_overrides=[
                TargetRule(
                    Target([TargetCondition(TargetKey('SEGMENT', "SEGMENT"),
                                            TargetMatch('MATCH', 'IN', 'STRING', ["seg_01"]))]),
                    TargetAction('BUCKET', None, 1)
                )
            ],
            target_audiences=[],
            target_rules=[],
            default_rule=TargetAction('BUCKET', bucket_id=1),
            container_id=None,
            winner_variation_id=None
        )

        self.target_matcher.matches.return_value = False

        request = create_experiment_request(experiment=experiment, user=user)

        # when
        actual = self.sut.resolve_or_none(request, Evaluator.context())

        # then
        self.assertIsNone(actual)



from hackle.internal.model.properties_builder import PropertiesBuilder
from tests import base


class PropertiesBuilderTest(base.BaseTest):

    def test_raw_value_valid_build(self):
        self.assertEqual(PropertiesBuilder().add("key1", 1).build(), {"key1": 1})
        self.assertEqual(PropertiesBuilder().add("key1", "1").build(), {"key1": "1"})
        self.assertEqual(PropertiesBuilder().add("key1", True).build(), {"key1": True})
        self.assertEqual(PropertiesBuilder().add("key1", False).build(), {"key1": False})

    def test_raw_value_invalid(self):
        self.assertEqual(PropertiesBuilder().add("key1", {id: "a"}).build(), {})

    def test_array_value(self):
        self.assertEqual(PropertiesBuilder().add("key1", [1, 2, 3]).build(), {"key1": [1, 2, 3]})
        self.assertEqual(PropertiesBuilder().add("key1", ["1", "2", "3"]).build(), {"key1": ["1", "2", "3"]})
        self.assertEqual(PropertiesBuilder().add("key1", ["1", 2, "3"]).build(), {"key1": ["1", 2, "3"]})
        self.assertEqual(PropertiesBuilder().add("key1", [1, 2, 3, None, 4]).build(), {"key1": [1, 2, 3, 4]})
        self.assertEqual(PropertiesBuilder().add("key1", [True, False]).build(), {"key1": []})
        self.assertEqual(PropertiesBuilder().add("key1", ["a" * 1024]).build(), {"key1": ["a" * 1024]})
        self.assertEqual(PropertiesBuilder().add("key1", ["a" * 1025]).build(), {"key1": []})

    def test_max_property_size_is_128(self):
        builder = PropertiesBuilder()
        for i in range(128):
            builder.add(str(i), i)

        self.assertEqual(128, len(builder.build()))

        builder.add("key", 42)
        properties = builder.build()
        self.assertEqual(128, len(properties))
        self.assertIsNone(properties.get("key"))

    def test_max_key_length_is_128(self):
        builder = PropertiesBuilder()

        builder.add("a" * 128, 128)
        self.assertEqual(1, len(builder.build()))

        builder.add("a" * 129, 1289)
        self.assertEqual(1, len(builder.build()))

    def test_properties(self):
        properties = {
            "k1": "v1",
            "k2": 2,
            "k3": True,
            "k4": False,
            "k5": [1, 2, 3],
            "k6": ["1", "2", "3"],
            "k7": None
        }

        actual = PropertiesBuilder().add_properties(properties).build()

        self.assertEqual(6, len(actual))
        self.assertEqual({
            "k1": "v1",
            "k2": 2,
            "k3": True,
            "k4": False,
            "k5": [1, 2, 3],
            "k6": ["1", "2", "3"],
        }, actual)

        self.assertEqual({}, PropertiesBuilder().add_properties(None).build())
        self.assertEqual({}, PropertiesBuilder().add_properties("str").build())



from unittest.mock import Mock

from hackle.internal.evaluation.flow.evaluation_flow_factory import EvaluationFlowFactory
from hackle.internal.evaluation.flow.flow_evaluator import *
from tests import base


class EvaluationFlowFactoryTest(base.BaseTest):

    def setUp(self):
        self.sut = EvaluationFlowFactory(Mock())

    def test_AB_TEST(self):
        flow = self.sut.get_evaluation_flow('AB_TEST')

        flow = self.is_decision_with(flow, OverrideEvaluator)
        flow = self.is_decision_with(flow, IdentifierEvaluator)
        flow = self.is_decision_with(flow, ContainerEvaluator)
        flow = self.is_decision_with(flow, ExperimentTargetEvaluator)
        flow = self.is_decision_with(flow, DraftEvaluator)
        flow = self.is_decision_with(flow, PausedEvaluator)
        flow = self.is_decision_with(flow, CompletedEvaluator)
        flow = self.is_decision_with(flow, TrafficAllocateEvaluator)
        self.assertTrue(flow.is_end)

    def test_FEATURE_FLAG(self):
        flow = self.sut.get_evaluation_flow('FEATURE_FLAG')

        flow = self.is_decision_with(flow, DraftEvaluator)
        flow = self.is_decision_with(flow, PausedEvaluator)
        flow = self.is_decision_with(flow, CompletedEvaluator)
        flow = self.is_decision_with(flow, OverrideEvaluator)
        flow = self.is_decision_with(flow, IdentifierEvaluator)
        flow = self.is_decision_with(flow, TargetRuleEvaluator)
        flow = self.is_decision_with(flow, DefaultRuleEvaluator)
        self.assertTrue(flow.is_end)

    def is_decision_with(self, evaluation_flow, cls):
        self.assertFalse(evaluation_flow.is_end)
        self.assertIsInstance(evaluation_flow.flow_evaluator, cls)
        return evaluation_flow.next_flow



from hackle.internal.evaluation.bucket.bucketer import Bucketer
from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.evaluation.evaluator.remoteconfig.remote_config_request import RemoteConfigRequest
from hackle.internal.evaluation.match.target_matcher import TargetMatcher
from hackle.internal.model.entities import RemoteConfigTargetRule


class RemoteConfigTargetRuleDeterminer(object):
    def __init__(self, matcher):
        """
        :param RemoteConfigTargetRuleMatcher matcher:
        """
        self.__matcher = matcher

    def determine_target_rule_or_none(self, request, context):
        """
        :param RemoteConfigRequest request:
        :param Evaluator.Context context:

        :rtype: RemoteConfigTargetRule or None
        """

        for target_rule in request.parameter.target_rules:
            if self.__matcher.matches(request, context, target_rule):
                return target_rule
        return None


class RemoteConfigTargetRuleMatcher(object):
    def __init__(self, target_matcher, bucketer):
        """
        :param TargetMatcher target_matcher:
        :param Bucketer bucketer:
        """
        self.__target_matcher = target_matcher
        self.__bucketer = bucketer

    def matches(self, request, context, target_rule):
        """
        :param RemoteConfigRequest request:
        :param Evaluator.Context context:
        :param RemoteConfigTargetRule target_rule:

        :rtype: bool
        """
        if not self.__target_matcher.matches(request, context, target_rule.target):
            return False

        identifier = request.user.identifiers.get(request.parameter.identifier_type)
        if identifier is None:
            return False

        bucket = request.workspace.get_bucket_or_none(target_rule.bucket_id)
        if bucket is None:
            raise Exception('Bucket[{}]'.format(target_rule.bucket_id))

        return self.__bucketer.bucketing(bucket, identifier) is not None



import json

from hackle.internal.event.user_event import UserEvent, TrackEvent, ExposureEvent, RemoteConfigEvent
from hackle.internal.http.http_client import HttpClient
from hackle.internal.logger.log import Log
from hackle.internal.user.identifier_type import IdentifierType


# noinspection PyMethodMayBeStatic
class EventSerializer(object):

    def serialize(self, events):
        """
        :param list[UserEvent] events:
        :rtype: str
        """
        return json.dumps(self.__payload(events))

    def __payload(self, events):
        """
        :param list[UserEvent] events:
        :rtype: dict
        """
        exposure_events = []
        track_events = []
        remote_config_events = []
        for event in events:
            if isinstance(event, ExposureEvent):
                exposure_events.append(self.__exposure_event(event))

            if isinstance(event, TrackEvent):
                track_events.append(self.__track_event(event))

            if isinstance(event, RemoteConfigEvent):
                remote_config_events.append(self.__remote_config_event(event))

        return {
            'exposureEvents': exposure_events,
            'trackEvents': track_events,
            'remoteConfigEvents': remote_config_events
        }

    def __exposure_event(self, event):
        """
        :param ExposureEvent event:
        :rtype: dict
        """
        return {
            "insertId": event.insert_id,
            "timestamp": event.timestamp,

            "userId": event.user.identifiers.get(IdentifierType.ID),
            "identifiers": event.user.identifiers,
            "userProperties": event.user.properties,
            "hackleProperties": {},

            "experimentId": event.experiment.id,
            "experimentKey": event.experiment.key,
            "experimentType": event.experiment.type,
            "experimentVersion": event.experiment.version,
            "variationId": event.variation_id,
            "variationKey": event.variation_key,
            "decisionReason": event.reason,
            "properties": event.properties
        }

    def __track_event(self, event):
        """
        :param  TrackEvent event:
        :rtype: dict
        """
        return {
            "insertId": event.insert_id,
            "timestamp": event.timestamp,

            "userId": event.user.identifiers.get(IdentifierType.ID),
            "identifiers": event.user.identifiers,
            "userProperties": event.user.properties,
            "hackleProperties": {},

            "eventTypeId": event.event_type.id,
            "eventTypeKey": event.event_type.key,
            "value": event.event.value,
            "properties": event.event.properties
        }

    def __remote_config_event(self, event):
        """
        :param RemoteConfigEvent event:
        :rtype: dict
        """
        return {
            "insertId": event.insert_id,
            "timestamp": event.timestamp,

            "userId": event.user.identifiers.get(IdentifierType.ID),
            "identifiers": event.user.identifiers,
            "userProperties": event.user.properties,
            "hackleProperties": {},

            "parameterId": event.parameter.id,
            "parameterKey": event.parameter.key,
            "parameterType": event.parameter.type,
            "valueId": event.value_id,
            "decisionReason": event.reason,
            "properties": event.properties
        }


class EventDispatcher(object):
    def __init__(self, base_event_url, http_client):
        """
        :param str base_event_url:
        :param HttpClient http_client:
        """
        self.__event_url = base_event_url + '/api/v2/events'
        self.__http_client = http_client
        self.__serializer = EventSerializer()

    def dispatch(self, events):
        """
        :param list[UserEvent] events:
        """

        try:
            dto = self.__serializer.serialize(events)
            self.__http_client.post(self.__event_url, dto)
        except Exception as e:
            Log.get().error('Failed to dispatch events: {}'.format(str(e)))



from unittest import TestCase
from unittest.mock import Mock

from hackle.decision import DecisionReason
from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.evaluation.evaluator.experiment.experiment_evaluation import ExperimentEvaluation
from hackle.internal.evaluation.evaluator.remoteconfig.remote_config_evaluation import RemoteConfigEvaluation
from hackle.internal.evaluation.match.experiment_condition_matcher import AbTestConditionMatcher
from hackle.internal.evaluation.match.value_operator_matcher import ValueOperatorMatcher
from hackle.internal.model.entities import TargetCondition, TargetKey, TargetMatch
from hackle.internal.workspace.workspace import Workspace
from tests.internal.evaluation.evaluator.experiment.test_experiment_request import create_experiment_request
from tests.internal.model.test_entities import create_experiment


class AbTestConditionMatcherTest(TestCase):

    def setUp(self):
        self.evaluator = Mock(spec=Evaluator)
        self.value_operator_matcher = Mock(spec=ValueOperatorMatcher)
        self.sut = AbTestConditionMatcher(self.evaluator, self.value_operator_matcher)
        self.context = Evaluator.context()

    def test_key_is_not_int(self):
        # given
        experiment = create_experiment(type='AB_TEST')
        request = create_experiment_request(experiment=experiment)
        condition = TargetCondition(
            TargetKey('AB_TEST', 'string'),
            TargetMatch('MATCH', 'IN', 'STRING', ['A'])
        )

        # when
        with self.assertRaises(Exception) as actual:
            self.sut.matches(request, Evaluator.context(), condition)

        # then
        self.assertEqual('Invalid key [AB_TEST, string]', str(actual.exception))

    def test_experiment_not_found(self):
        # given
        experiment = create_experiment(type='AB_TEST')
        request = create_experiment_request(experiment=experiment)
        condition = TargetCondition(
            TargetKey('AB_TEST', '42'),
            TargetMatch('MATCH', 'IN', 'STRING', ['A'])
        )

        # when
        actual = self.sut.matches(request, Evaluator.context(), condition)

        # then
        self.assertFalse(actual)

    # noinspection PyMethodMayBeStatic
    def __request(self, experiment):
        workspace = Mock(spec=Workspace)
        workspace.get_experiment_or_none.return_value = experiment
        return create_experiment_request(workspace=workspace, experiment=experiment)

    def test_not_matched_reason(self):
        def check(reason):
            # given
            experiment = create_experiment(type='AB_TEST')
            request = self.__request(experiment=experiment)
            condition = TargetCondition(
                TargetKey('AB_TEST', '42'),
                TargetMatch('MATCH', 'IN', 'STRING', ['A'])
            )

            evaluation = ExperimentEvaluation.of(request, self.context, request.experiment.variations[0], reason)
            self.evaluator.evaluate.return_value = evaluation

            # when
            actual = self.sut.matches(request, Evaluator.context(), condition)

            # then
            self.assertFalse(actual)

        check(DecisionReason.EXPERIMENT_DRAFT)
        check(DecisionReason.EXPERIMENT_PAUSED)
        check(DecisionReason.NOT_IN_MUTUAL_EXCLUSION_EXPERIMENT)
        check(DecisionReason.VARIATION_DROPPED)
        check(DecisionReason.NOT_IN_EXPERIMENT_TARGET)
        self.value_operator_matcher.matches.assert_not_called()

    def test_matched_reason(self):
        def check(reason):
            # given
            experiment = create_experiment(type='AB_TEST')
            request = self.__request(experiment=experiment)
            condition = TargetCondition(
                TargetKey('AB_TEST', '42'),
                TargetMatch('MATCH', 'IN', 'STRING', ['A'])
            )

            evaluation = ExperimentEvaluation.of(request, self.context, request.experiment.variations[0], reason)
            self.evaluator.evaluate.return_value = evaluation
            self.value_operator_matcher.matches.return_value = True

            # when
            actual = self.sut.matches(request, Evaluator.context(), condition)

            # then
            self.assertTrue(actual)

        check(DecisionReason.OVERRIDDEN)
        check(DecisionReason.TRAFFIC_ALLOCATED)
        check(DecisionReason.TRAFFIC_ALLOCATED_BY_TARGETING)
        check(DecisionReason.EXPERIMENT_COMPLETED)

    def test_already_evaluated_experiment(self):
        # given
        experiment = create_experiment(type='AB_TEST')
        request = self.__request(experiment=experiment)
        condition = TargetCondition(
            TargetKey('AB_TEST', '42'),
            TargetMatch('MATCH', 'IN', 'STRING', ['A'])
        )
        evaluation = ExperimentEvaluation.of(request, self.context, request.experiment.variations[0],
                                             DecisionReason.TRAFFIC_ALLOCATED)
        self.evaluator.evaluate.return_value = evaluation
        self.value_operator_matcher.matches.return_value = True

        self.context.add_evaluation(evaluation)

        # when
        actual = self.sut.matches(request, self.context, condition)

        # then
        self.assertTrue(actual)
        self.evaluator.evaluate.assert_not_called()
        self.assertEqual(1, len(self.context.target_evaluations))

    def test_replace_reason(self):
        # given
        experiment = create_experiment(type='AB_TEST')
        request = self.__request(experiment=experiment)
        condition = TargetCondition(
            TargetKey('AB_TEST', '42'),
            TargetMatch('MATCH', 'IN', 'STRING', ['A'])
        )
        evaluation = ExperimentEvaluation.of(request, self.context, request.experiment.variations[0],
                                             DecisionReason.TRAFFIC_ALLOCATED)
        self.evaluator.evaluate.return_value = evaluation
        self.value_operator_matcher.matches.return_value = True

        # when
        actual = self.sut.matches(request, self.context, condition)

        # then
        self.assertTrue(actual)
        self.assertEqual(DecisionReason.TRAFFIC_ALLOCATED_BY_TARGETING,
                         self.context.get_evaluation_or_none(experiment).reason)

    def test_use_evaluation_directly(self):
        # given
        experiment = create_experiment(type='AB_TEST')
        request = self.__request(experiment=experiment)
        condition = TargetCondition(
            TargetKey('AB_TEST', '42'),
            TargetMatch('MATCH', 'IN', 'STRING', ['A'])
        )
        evaluation = ExperimentEvaluation.of(request, self.context, request.experiment.variations[0],
                                             DecisionReason.OVERRIDDEN)
        self.evaluator.evaluate.return_value = evaluation
        self.value_operator_matcher.matches.return_value = True

        # when
        actual = self.sut.matches(request, self.context, condition)

        # then
        self.assertTrue(actual)
        self.assertEqual(evaluation, self.context.get_evaluation_or_none(experiment))

    def test_unexpected_evaluation(self):
        # given
        experiment = create_experiment(type='AB_TEST')
        request = self.__request(experiment=experiment)
        condition = TargetCondition(
            TargetKey('AB_TEST', '42'),
            TargetMatch('MATCH', 'IN', 'STRING', ['A'])
        )
        evaluation = Mock(spec=RemoteConfigEvaluation)
        self.evaluator.evaluate.return_value = evaluation
        self.value_operator_matcher.matches.return_value = True

        # when
        with self.assertRaises(Exception):
            self.sut.matches(request, self.context, condition)



class ExperimentDecision:
    def __init__(self, variation, reason, config=None):
        if config is None:
            config = {}
        self.variation = variation
        self.reason = reason
        self.config = config

    def __eq__(self, o):
        return self.variation == o.variation and self.reason == o.reason

    def __ne__(self, o):
        return not self.__eq__(o)

    def __str__(self):
        return '(variation={}, reason={}, config={})'.format(self.variation, self.reason, self.config)

    def __repr__(self):
        return self.__str__()

    def get(self, key, default=None):
        return self.config.get(key, default)


class FeatureFlagDecision:
    def __init__(self, is_on, reason, config=None):
        if config is None:
            config = {}
        self.is_on = is_on
        self.reason = reason
        self.config = config

    def __eq__(self, o):
        return self.is_on == o.is_on and self.reason == o.reason

    def __ne__(self, o):
        return not self.__eq__(o)

    def __str__(self):
        return '(is_on={}, reason={}, config={})'.format(self.is_on, self.reason, self.config)

    def __repr__(self):
        return self.__str__()

    def get(self, key, default=None):
        return self.config.get(key, default)


class RemoteConfigDecision:
    def __init__(self, value, reason):
        self.value = value
        self.reason = reason

    def __eq__(self, o):
        return self.value == o.value and self.reason == o.reason

    def __ne__(self, o):
        return not self.__eq__(o)

    def __str__(self):
        return '(value={}, reason={})'.format(self.value, self.reason)

    def __repr__(self):
        return self.__str__()


class DecisionReason(object):
    SDK_NOK_READY = 'SDK_NOK_READY'
    EXCEPTION = 'EXCEPTION'
    INVALID_INPUT = 'INVALID_INPUT'

    EXPERIMENT_NOT_FOUND = 'EXPERIMENT_NOT_FOUND'
    EXPERIMENT_DRAFT = 'EXPERIMENT_DRAFT'
    EXPERIMENT_PAUSED = 'EXPERIMENT_PAUSED'
    EXPERIMENT_COMPLETED = 'EXPERIMENT_COMPLETED'
    OVERRIDDEN = 'OVERRIDDEN'
    TRAFFIC_NOT_ALLOCATED = 'TRAFFIC_NOT_ALLOCATED'
    TRAFFIC_ALLOCATED = 'TRAFFIC_ALLOCATED'
    TRAFFIC_ALLOCATED_BY_TARGETING = 'TRAFFIC_ALLOCATED_BY_TARGETING'
    VARIATION_DROPPED = 'VARIATION_DROPPED'
    NOT_IN_EXPERIMENT_TARGET = 'NOT_IN_EXPERIMENT_TARGET'
    NOT_IN_MUTUAL_EXCLUSION_EXPERIMENT = 'NOT_IN_MUTUAL_EXCLUSION_EXPERIMENT'
    IDENTIFIER_NOT_FOUND = 'IDENTIFIER_NOT_FOUND'

    FEATURE_FLAG_NOT_FOUND = 'FEATURE_FLAG_NOT_FOUND'
    FEATURE_FLAG_INACTIVE = 'FEATURE_FLAG_INACTIVE'
    INDIVIDUAL_TARGET_MATCH = 'INDIVIDUAL_TARGET_MATCH'
    TARGET_RULE_MATCH = 'TARGET_RULE_MATCH'
    DEFAULT_RULE = 'DEFAULT_RULE'

    REMOTE_CONFIG_PARAMETER_NOT_FOUND = 'REMOTE_CONFIG_PARAMETER_NOT_FOUND'
    TYPE_MISMATCH = 'TYPE_MISMATCH'



from unittest import TestCase

from hackle.decision import DecisionReason
from hackle.internal.event.user_event import RemoteConfigEvent, ExposureEvent
from hackle.internal.hackle_core import HackleCore
from hackle.internal.user.identifier_type import IdentifierType
from hackle.internal.user.internal_hackle_user import InternalHackleUser
from tests.internal.event.test_in_memory_event_processor import InMemoryEventProcessor
from tests.internal.workspace.test_resource_workspace_fetcher import ResourceWorkspaceFetcher


class HackleCoreTest2(TestCase):

    #
    #      RC(1)
    #     /     \
    #    /       \
    # AB(2)     FF(4)
    #   |   \     |
    #   |     \   |
    # AB(3)     FF(5)
    #             |
    #             |
    #           AB(6)
    #
    def test_target_experiment(self):
        workspace_fetcher = ResourceWorkspaceFetcher('target_experiment.json')
        event_processor = InMemoryEventProcessor()
        # noinspection PyTypeChecker
        core = HackleCore.create(workspace_fetcher, event_processor)

        user = InternalHackleUser.builder().identifier(IdentifierType.ID, "user").build()
        decision = core.remote_config('rc', user, 'STRING', '42')

        self.assertEqual('Targeting!!', decision.value)
        self.assertEqual(DecisionReason.TARGET_RULE_MATCH, decision.reason)

        event = event_processor.processed_events[0]
        self.assertIsInstance(event, RemoteConfigEvent)
        self.assertEqual(
            {
                "requestValueType": "STRING",
                "requestDefaultValue": "42",
                "targetRuleKey": "rc_1_key",
                "targetRuleName": "rc_1_name",
                "returnValue": "Targeting!!"
            },
            event.properties
        )

        for event in event_processor.processed_events[1:]:
            self.assertIsInstance(event, ExposureEvent)
            self.assertEqual(
                {
                    "$targetingRootType": "REMOTE_CONFIG",
                    "$targetingRootId": 1
                },
                event.properties
            )

    #
    #     RC(1)
    #      ↓
    # ┌── AB(2)
    # ↑    ↓
    # |   FF(3)
    # ↑    ↓
    # |   AB(4)
    # └────┘
    #
    def test_experiment_circular(self):
        workspace_fetcher = ResourceWorkspaceFetcher('target_experiment_circular.json')
        event_processor = InMemoryEventProcessor()
        # noinspection PyTypeChecker
        core = HackleCore.create(workspace_fetcher, event_processor)

        user = InternalHackleUser.builder().identifier(IdentifierType.ID, "user").build()

        with self.assertRaises(Exception) as actual:
            core.remote_config('rc', user, 'STRING', 'XXX')

        self.assertIn('Circular evaluation has occurred', str(actual.exception))

    #
    #                     Container(1)
    # ┌──────────────┬───────────────────────────────────────┐
    # | ┌──────────┐ |                                       |
    # | |   AB(2)  | |                                       |
    # | └──────────┘ |                                       |
    # └──────────────┴───────────────────────────────────────┘
    #       25 %                        75 %
    #
    def test_container(self):
        workspace_fetcher = ResourceWorkspaceFetcher('container.json')
        event_processor = InMemoryEventProcessor()
        # noinspection PyTypeChecker
        core = HackleCore.create(workspace_fetcher, event_processor)

        decisions = []
        for i in range(0, 10000):
            user = InternalHackleUser.builder().identifier(IdentifierType.ID, str(i)).build()
            decisions.append(core.experiment(2, user, 'A'))

        self.assertEqual(10000, len(event_processor.processed_events))
        self.assertEqual(10000, len(decisions))
        self.assertEqual(2452, len(list(filter(lambda it: it.reason == 'TRAFFIC_ALLOCATED', decisions))))
        self.assertEqual(7548,
                         len(list(filter(lambda it: it.reason == 'NOT_IN_MUTUAL_EXCLUSION_EXPERIMENT', decisions))))

    def test_segment_match(self):
        workspace_fetcher = ResourceWorkspaceFetcher('segment_match.json')
        event_processor = InMemoryEventProcessor()
        # noinspection PyTypeChecker
        core = HackleCore.create(workspace_fetcher, event_processor)

        user_1 = InternalHackleUser.builder().identifier(IdentifierType.ID, "matched_id").build()
        decision_1 = core.experiment(1, user_1, 'A')
        self.assertEqual('A', decision_1.variation)
        self.assertEqual(DecisionReason.OVERRIDDEN, decision_1.reason)

        user_2 = InternalHackleUser.builder().identifier(IdentifierType.ID, "not_matched_id").build()
        decision_2 = core.experiment(1, user_2, 'A')
        self.assertEqual('A', decision_2.variation)
        self.assertEqual(DecisionReason.TRAFFIC_ALLOCATED, decision_2.reason)



from hackle.decision import DecisionReason
from hackle.internal.evaluation.evaluator.evaluator import ContextualEvaluator, Evaluator
from hackle.internal.evaluation.evaluator.remoteconfig.remote_config_evaluation import RemoteConfigEvaluation
from hackle.internal.evaluation.evaluator.remoteconfig.remote_config_request import RemoteConfigRequest
from hackle.internal.evaluation.target.remote_config_target_rule_determiner import RemoteConfigTargetRuleDeterminer
from hackle.internal.model.entities import RemoteConfigParameterValue
from hackle.internal.model.properties_builder import PropertiesBuilder
from hackle.internal.type import hackle_types


class RemoteConfigEvaluator(ContextualEvaluator):

    def __init__(self, target_rule_determiner):
        """
        :param RemoteConfigTargetRuleDeterminer target_rule_determiner:
        """
        super(RemoteConfigEvaluator, self).__init__()
        self.__target_rule_determiner = target_rule_determiner

    def supports(self, request):
        return isinstance(request, RemoteConfigRequest)

    def evaluate_internal(self, request, context):
        """
        :param RemoteConfigRequest request:
        :param Evaluator.Context context:

        :rtype: RemoteConfigEvaluation
        """
        properties_builder = PropertiesBuilder() \
            .add("requestValueType", request.required_type) \
            .add("requestDefaultValue", request.default_value)

        if request.user.identifiers.get(request.parameter.identifier_type) is None:
            return RemoteConfigEvaluation.of_default(
                request, context, DecisionReason.IDENTIFIER_NOT_FOUND, properties_builder)

        target_rule = self.__target_rule_determiner.determine_target_rule_or_none(request, context)
        if target_rule is not None:
            properties_builder.add("targetRuleKey", target_rule.key)
            properties_builder.add("targetRuleName", target_rule.name)
            return self.__evaluation(
                request, context, target_rule.value, DecisionReason.TARGET_RULE_MATCH, properties_builder)

        return self.__evaluation(
            request, context, request.parameter.default_value, DecisionReason.DEFAULT_RULE, properties_builder)

    def __evaluation(self, request, context, parameter_value, reason, properties_builder):
        """
        :param RemoteConfigRequest request:
        :param Evaluator.Context context:
        :param RemoteConfigParameterValue parameter_value:
        :param str reason:
        :param PropertiesBuilder properties_builder:

        :rtype: RemoteConfigEvaluation
        """
        if self.__is_valid(request.required_type, parameter_value.raw_value):
            return RemoteConfigEvaluation.of(
                request,
                context,
                parameter_value.id,
                parameter_value.raw_value,
                reason,
                properties_builder
            )
        else:
            return RemoteConfigEvaluation.of_default(
                request,
                context,
                DecisionReason.TYPE_MISMATCH,
                properties_builder
            )

    # noinspection PyMethodMayBeStatic
    def __is_valid(self, required_type, value):
        if required_type == 'NULL':
            return True
        elif required_type == 'STRING':
            return hackle_types.is_string(value)
        elif required_type == 'NUMBER':
            return hackle_types.is_number(value)
        elif required_type == 'BOOLEAN':
            return hackle_types.is_bool(value)
        else:
            return False



from hackle.internal.evaluation.match.condition_matcher import ConditionMatcher
from hackle.internal.evaluation.match.user_condition_matcher import UserConditionMatcher
from hackle.internal.logger.log import Log
from hackle.internal.model.entities import Segment, Target
from hackle.internal.type import hackle_types


class SegmentConditionMatcher(ConditionMatcher):

    def __init__(self, segment_matcher):
        """
        :param SegmentMatcher segment_matcher:
        """
        self.__segment_matcher = segment_matcher

    def matches(self, request, context, condition):
        if condition.key.type != 'SEGMENT':
            raise Exception('Unsupported target.key.type [{}]'.format(condition.key.type))

        for value in condition.match.values:
            if self.__matches(request, context, value):
                return self.__match_type(condition.match.type, True)

        return self.__match_type(condition.match.type, False)

    def __matches(self, request, context, value):
        """
        :param Evaluator.Request request:
        :param Evaluator.Context context:
        :param object value:
        :rtype: bool
        """
        segment_key = value
        if not hackle_types.is_string(segment_key):
            raise Exception('SegmentKey[{}]'.format(segment_key))
        segment = request.workspace.get_segment_or_none(segment_key)
        if segment is None:
            raise Exception('Segment[{}]'.format(segment_key))

        return self.__segment_matcher.matches(request, context, segment)

    # noinspection PyMethodMayBeStatic
    def __match_type(self, match_type, is_matched):
        """
        :param str match_type:
        :param bool is_matched:

        :rtype: bool
        """
        if match_type == 'MATCH':
            return is_matched
        elif match_type == 'NOT_MATCH':
            return not is_matched
        else:
            Log.get().debug('Unsupported type[{}]. Please use the latest version of sdk.'.format(match_type))
            return False


class SegmentMatcher(object):
    def __init__(self, user_condition_matcher):
        """
        :param UserConditionMatcher user_condition_matcher:
        """
        self.__user_condition_matcher = user_condition_matcher

    def matches(self, request, context, segment):
        """
        :param Evaluator.Request request:
        :param Evaluator.Context context:
        :param Segment segment:

        :rtype: bool
        """
        for target in segment.targets:
            if self.__matches(request, context, target):
                return True
        return False

    def __matches(self, request, context, target):
        """
        :param Evaluator.Request request:
        :param Evaluator.Context context:
        :param Target target:

        :rtype: bool
        """
        for condition in target.conditions:
            if not self.__user_condition_matcher.matches(request, context, condition):
                return False
        return True



import abc

from six import add_metaclass

from .internal.logger.log import Log
from .internal.model.properties_builder import PropertiesBuilder
from .internal.type import hackle_types
from .internal.user.identifiers_builder import IdentifiersBuilder


class User(object):
    def __init__(self, id, properties):
        self.id = id
        self.properties = properties

    def __str__(self):
        return 'User(id={}, properties={})'.format(self.id, self.properties)


class HackleUser(object):
    def __init__(
            self,
            id=None,
            user_id=None,
            device_id=None,
            identifiers=None,
            properties=None):
        self.__id = id
        self.__user_id = user_id
        self.__device_id = device_id
        self.__identifiers = identifiers
        self.__properties = properties

    @staticmethod
    def builder():
        return HackleUserBuilder()

    @staticmethod
    def of(user):
        return HackleUser.builder().id(user.id).properties(user.properties).build()

    @property
    def id(self):
        """
        :rtype: string or None
        """
        return self.__id

    @property
    def user_id(self):
        """
        :rtype: string or None
        """
        return self.__user_id

    @property
    def device_id(self):
        """
        :rtype: string or None
        """
        return self.__device_id

    @property
    def identifiers(self):
        """
        :rtype: dict[str, str]
        """
        return self.__identifiers

    @property
    def properties(self):
        """
        :rtype: dict[str, object]
        """
        return self.__properties

    def __str__(self):
        return 'HackleUser(id={}, user_id={}, device_id={}, identifiers={}, properties={})'.format(self.__id,
                                                                                                   self.__user_id,
                                                                                                   self.__device_id,
                                                                                                   self.__identifiers,
                                                                                                   self.__properties)

    def __repr__(self):
        return self.__str__()


class HackleUserBuilder:

    def __init__(self):
        self.__id = None
        self.__user_id = None
        self.__device_id = None
        self.__identifiers = IdentifiersBuilder()
        self.__properties = PropertiesBuilder()

    def id(self, id_):
        self.__id = IdentifiersBuilder.sanitize_value_or_none(id_)
        return self

    def user_id(self, user_id):
        self.__user_id = IdentifiersBuilder.sanitize_value_or_none(user_id)
        return self

    def device_id(self, device_id):
        self.__device_id = IdentifiersBuilder.sanitize_value_or_none(device_id)
        return self

    def session_id(self, session_id):
        return self.identifier("$sessionId", session_id)

    def identifier(self, identifier_type, identifier_value):
        self.__identifiers.add(identifier_type, identifier_value)
        return self

    def identifiers(self, identifiers):
        self.__identifiers.add_identifiers(identifiers)
        return self

    def property(self, property_key, property_value):
        self.__properties.add(property_key, property_value)
        return self

    def properties(self, properties):
        self.__properties.add_properties(properties)
        return self

    def build(self):
        return HackleUser(
            self.__id,
            self.__user_id,
            self.__device_id,
            self.__identifiers.build(),
            self.__properties.build(),
        )


class Event(object):
    def __init__(self, key, value, properties):
        self.key = key
        self.value = value
        self.properties = properties


class HackleEvent(object):
    def __init__(self, key, value, properties):
        """
        :param str key:
        :param float or None value:
        :param dict properties:
        """
        self.__key = key
        self.__value = value
        self.__properties = properties

    @staticmethod
    def builder(key):
        return HackleEventBuilder(key)

    @staticmethod
    def from_event(event):
        if isinstance(event, HackleEvent):
            return event

        return HackleEvent.builder(event.key) \
            .value(event.value) \
            .properties(event.properties) \
            .build()

    @property
    def key(self):
        """
        :rtype: str
        """
        return self.__key

    @property
    def value(self):
        """
        :rtype: float or None
        """
        return self.__value

    @property
    def properties(self):
        """
        :rtype: dict[str, object]
        """
        return self.__properties

    @property
    def is_valid(self):
        """
        :rtype: bool
        """
        return self.error_or_none is None

    @property
    def error_or_none(self):
        """
        :rtype: str or None
        """
        if not hackle_types.is_not_empty_string(self.__key):
            return "Invalid event key: {} (expected: not empty string)".format(self.__key)

        if self.__value is not None and not hackle_types.is_finite_number(self.__value):
            return "Invalid event value: {} (expected: finite number)"

        if self.__properties is None or not isinstance(self.__properties, dict):
            return "Invalid event properties: {} (expected: dict)".format(self.__properties)

        return None

    def __eq__(self, other):
        if not isinstance(other, HackleEvent):
            return False
        return self.__key == other.__key and self.__value == other.__value and self.__properties == other.__properties

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        if not self.is_valid:
            return "InvalidHackleEvent({})".format(self.error_or_none)

        return "HackleEvent(key={}, value={}, properties={})".format(self.__key, self.__value, self.__properties)

    def __repr__(self):
        return self.__str__()


class HackleEventBuilder(object):

    def __init__(self, key):
        self.__key = key
        self.__value = None
        self.__properties = PropertiesBuilder()

    def build(self):
        return HackleEvent(
            self.__key,
            self.__value,
            self.__properties.build()
        )

    def value(self, value):
        self.__value = value
        return self

    def property(self, key, value):
        self.__properties.add(key, value)
        return self

    def properties(self, properties):
        self.__properties.add_properties(properties)
        return self


class PropertyOperation(object):
    SET = '$set'
    SET_ONCE = '$setOnce'
    UNSET = '$unset'
    INCREMENT = '$increment'
    APPEND = '$append'
    APPEND_ONCE = '$appendOnce'
    PREPEND = '$prepend'
    PREPEND_ONCE = '$prependOnce'
    REMOVE = '$remove'
    CLEAR_ALL = '$clearAll'


class PropertyOperations(object):

    def __init__(self, operations):
        """
        :param dict[str, dict[str, object]] operations:
        """
        self.__operations = operations

    @staticmethod
    def builder():
        """
        :rtype:  PropertyOperationsBuilder
        """
        return PropertyOperationsBuilder()

    def to_event(self):
        """
        :rtype: HackleEvent
        """
        builder = HackleEvent.builder('$properties')
        for operation in self.__operations:
            builder.property(operation, self.__operations[operation])
        return builder.build()

    def __iter__(self):
        for operation in self.__operations:
            yield operation, self.__operations[operation]

    def __str__(self):
        return 'PropertyOperations({})'.format(self.__operations.__str__())

    def __repr__(self):
        return self.__str__()


class PropertyOperationsBuilder(object):

    def __init__(self):
        self.__operations = {}  # type: dict[str, PropertiesBuilder]

    def set(self, key, value):
        """
        :param str key:
        :param object value:
        :rtype: PropertyOperationsBuilder
        """
        self.__add(PropertyOperation.SET, key, value)
        return self

    def set_once(self, key, value):
        """
        :param str key:
        :param object value:
        :rtype: PropertyOperationsBuilder
        """
        self.__add(PropertyOperation.SET_ONCE, key, value)
        return self

    def unset(self, key):
        """
        :param str key:
        :rtype: PropertyOperationsBuilder
        """
        self.__add(PropertyOperation.UNSET, key, '-')
        return self

    def increment(self, key, value):
        """
        :param str key:
        :param float value:
        :rtype: PropertyOperationsBuilder
        """
        self.__add(PropertyOperation.INCREMENT, key, value)
        return self

    def append(self, key, value):
        """
        :param str key:
        :param object value:
        :rtype: PropertyOperationsBuilder
        """
        self.__add(PropertyOperation.APPEND, key, value)
        return self

    def append_once(self, key, value):
        """
        :param str key:
        :param object value:
        :rtype: PropertyOperationsBuilder
        """
        self.__add(PropertyOperation.APPEND_ONCE, key, value)
        return self

    def prepend(self, key, value):
        """
        :param str key:
        :param object value:
        :rtype: PropertyOperationsBuilder
        """
        self.__add(PropertyOperation.PREPEND, key, value)
        return self

    def prepend_once(self, key, value):
        """
        :param str key:
        :param object value:
        :rtype: PropertyOperationsBuilder
        """
        self.__add(PropertyOperation.PREPEND_ONCE, key, value)
        return self

    def remove(self, key, value):
        """
        :param str key:
        :param object value:
        :rtype: PropertyOperationsBuilder
        """
        self.__add(PropertyOperation.REMOVE, key, value)
        return self

    def clear_all(self):
        """
        :rtype: PropertyOperationsBuilder
        """
        self.__add(PropertyOperation.CLEAR_ALL, 'clearAll', '-')
        return self

    def __add(self, operation, key, value):
        """
        :param str operation:
        :param str key:
        :param object or None value:
        """
        if self.__contains__(key):
            Log.get().debug('Property already added. Ignore the operation. [operation={}, key={}, value={}]'
                      .format(operation, key, value))
            return

        builder = self.__operations.get(operation) or self.__operations.setdefault(operation, PropertiesBuilder())
        builder.add(key, value)

    def build(self):
        """
        :rtype: PropertyOperations
        """
        operations = {}
        for operation in self.__operations:
            operations[operation] = self.__operations[operation].build()
        return PropertyOperations(operations)

    def __contains__(self, item):
        for operation in self.__operations:
            if item in self.__operations[operation]:
                return True
        return False


class Hackle:
    @staticmethod
    def user(id, **kwargs):
        return User(id, kwargs)

    @staticmethod
    def event(key, value=None, **kwargs):
        return Event(key, value, kwargs)


@add_metaclass(abc.ABCMeta)
class HackleRemoteConfig(object):
    @abc.abstractmethod
    def get(self, key, default=None):
        pass



import logging
import warnings

from .commons import enums

_DEFAULT_LOG_FORMAT = '%(levelname)-8s %(threadName)s %(asctime)s %(filename)s:%(lineno)s:%(message)s'


def reset_logger(name, level=None, handler=None):
    if level is None:
        level = logging.INFO
    logger = logging.getLogger(name)
    logger.setLevel(level)

    handler = handler or logging.StreamHandler()
    handler.setFormatter(logging.Formatter(_DEFAULT_LOG_FORMAT))

    logger.handlers = [handler]
    return logger


class BaseLogger(object):
    @staticmethod
    def log(*args):
        pass


class NoOpLogger(BaseLogger):
    def __init__(self):
        self.logger = reset_logger(
            name='.'.join([__name__, self.__class__.__name__]),
            level=logging.NOTSET,
            handler=logging.NullHandler(),
        )


class SimpleLogger(BaseLogger):
    def __init__(self, min_level=enums.LogLevels.INFO):
        self.level = min_level
        self.logger = reset_logger(name='.'.join([__name__, self.__class__.__name__]), level=min_level)

    def log(self, log_level, message):
        warnings.warn('this log is deprecated', DeprecationWarning)
        self.logger.log(log_level, message)


def adapt_logger(logger):
    if isinstance(logger, logging.Logger):
        return logger

    if isinstance(logger, (SimpleLogger, NoOpLogger)):
        return logger.logger

    return logger



import abc

from six import add_metaclass

from hackle.internal.time.time_unit import TimeUnit


@add_metaclass(abc.ABCMeta)
class Scheduler(object):

    @abc.abstractmethod
    def schedule_periodically(self, delay, period, unit, task):
        """
        :param float delay:
        :param float period:
        :param TimeUnit unit:
        :param function task:
        :rtype: ScheduledJob
        """
        pass

    @abc.abstractmethod
    def close(self):
        pass


@add_metaclass(abc.ABCMeta)
class ScheduledJob(object):
    @abc.abstractmethod
    def cancel(self):
        pass



from hackle.internal.evaluation.match.semantic_version import SemanticVersion
from tests import base


class SemanticVersionTest(base.BaseTest):

    def test_parse_string_타입이_아니면_None_리턴(self):
        self.assertIsNone(self.v(42))

    def test_parse_invalid_format(self):
        self.assertIsNone(self.v('01.0.0'))
        self.assertIsNone(self.v('1.01.0'))
        self.assertIsNone(self.v('1.1.01'))
        self.assertIsNone(self.v('2.x'))
        self.assertIsNone(self.v('2.3.x'))
        self.assertIsNone(self.v('2.3.1.4'))
        self.assertIsNone(self.v('2.3.1*beta'))
        self.assertIsNone(self.v('2.3.1-beta*'))
        self.assertIsNone(self.v('2.4.1-beta_4'))

    def test_parse_semantic_core_version(self):
        self.verify('1.0.0', 1, 0, 0, [], [])
        self.verify('14.165.14', 14, 165, 14, [], [])

    def test_parse_semantic_version_with_prerelease(self):
        self.verify('1.0.0-beta1', 1, 0, 0, ['beta1'], [])
        self.verify('1.0.0-beta.1', 1, 0, 0, ['beta', '1'], [])
        self.verify('1.0.0-x.y.z', 1, 0, 0, ['x', 'y', 'z'], [])

    def test_parse_semantic_version_with_build(self):
        self.verify('1.0.0+build1', 1, 0, 0, [], ['build1'])
        self.verify('1.0.0+build.1', 1, 0, 0, [], ['build', '1'])
        self.verify('1.0.0+1.2.3', 1, 0, 0, [], ['1', '2', '3'])

    def test_parse_semantic_version_with_prerelease_and_build(self):
        self.verify('1.0.0-alpha.3.rc.5+build.53', 1, 0, 0, ['alpha', '3', 'rc', '5'], ['build', '53'])

    def test_minor_patch가_없는경우_0으로_채워준다(self):
        self.verify('15', 15, 0, 0, [], [])
        self.verify('15.143', 15, 143, 0, [], [])
        self.verify('15-x.y.z', 15, 0, 0, ['x', 'y', 'z'], [])
        self.verify('15-x.y.z+a.b.c', 15, 0, 0, ['x', 'y', 'z'], ['a', 'b', 'c'])

    def test_core_버전만_있는경우_core_버전이_같으면_같은_버전이다(self):
        v1 = self.v('2.3.4')
        v2 = self.v('2.3.4')
        self.assertTrue(v1 == v2)

    def test_core_prerelease_버전이_모두_같아야_같은_버전이다(self):
        self.assertTrue(self.v('2.3.4-beta.1') == self.v('2.3.4-beta.1'))

    def test_prerelease_버전이_다르면_다른_버전이다(self):
        self.assertTrue(self.v('2.3.4-beta.1') != self.v('2.3.4-beta.2'))

    def test_build가_달라도_나머지가_같으면_같은버전이다(self):
        self.assertTrue(self.v('2.3.4+build.111') == self.v('2.3.4+build.222'))
        self.assertTrue(self.v('2.3.4-beta.1+build.111') == self.v('2.3.4-beta.1+build.222'))

    def test_major를_가장_먼저_비교한다(self):
        self.assertTrue(self.v('4.5.7') > self.v('3.5.7'))
        self.assertTrue(self.v('2.5.7') < self.v('3.5.7'))

    def test_major가_같으면_minor를_비교한다(self):
        self.assertTrue(self.v('3.6.7') > self.v('3.5.7'))
        self.assertTrue(self.v('3.4.7') < self.v('3.5.7'))

    def test_minor까지_같으면_patch를_비교한다(self):
        self.assertTrue(self.v('3.5.8') > self.v('3.5.7'))
        self.assertTrue(self.v('3.5.6') < self.v('3.5.7'))

    def test_정식_배포_버전이_더_높은_버전이다(self):
        self.assertTrue(self.v('3.5.7') > self.v('3.5.7-beta'))
        self.assertTrue(self.v('3.5.7-alpha') < self.v('3.5.7'))

    def test_prerelease_숫자로만_구성된_식별자는_수의_크기로_비교한다(self):
        self.assertTrue(self.v('3.5.7-1') < self.v('3.5.7-2'))
        self.assertTrue(self.v('3.5.7-1.1') < self.v('3.5.7-1.2'))
        self.assertTrue(self.v('3.5.7-11') > self.v('3.5.7-2'))

    def test_prerelease_알파벳이_포함된_경우에는_아스키_문자열_정렬을_한다(self):
        self.assertTrue(self.v('3.5.7-a') == self.v('3.5.7-a'))
        self.assertTrue(self.v('3.5.7-a') < self.v('3.5.7-b'))
        self.assertTrue(self.v('3.5.7-az') > self.v('3.5.7-ab'))
        self.assertTrue(self.v('3.5.7-alpha') < self.v('3.5.7-beta'))

    def test_prerelease_숫자로만_구성된_식별자는_어떤_경우에도_문자와_붙임표가_있는_식별자보다_낮은_우선순위로_여긴다(self):
        self.assertTrue(self.v('3.5.7-9') < self.v('3.5.7-a'))
        self.assertTrue(self.v('3.5.7-9') < self.v('3.5.7-a-9'))
        self.assertTrue(self.v('3.5.7-beta') > self.v('3.5.7-1'))

    def test_prerelease_앞선_식별자가_모두_같은_배포_전_버전의_경우에는_필드_수가_많은_쪽이_더_높은_우선순위를_가진다(self):
        self.assertTrue(self.v('3.5.7-alpha') < self.v('3.5.7-alpha.1'))
        self.assertTrue(self.v('3.5.7-1.2.3') < self.v('3.5.7-1.2.3.4'))

    def verify(self, version, major, minor, patch, prerelease, build):
        parsed_version = self.v(version)
        self.assertIsNotNone(parsed_version)
        self.assertEqual(major, parsed_version.core_version.major)
        self.assertEqual(minor, parsed_version.core_version.minor)
        self.assertEqual(patch, parsed_version.core_version.patch)
        self.assertEqual(prerelease, parsed_version.prerelease.identifiers)
        self.assertEqual(build, parsed_version.build.identifiers)

    @staticmethod
    def v(version):
        return SemanticVersion.parse_or_none(version)



from hackle.internal.evaluation.evaluator.evaluator import Evaluator, ContextualEvaluator


class DelegatingEvaluator(Evaluator):

    def __init__(self):
        super(DelegatingEvaluator, self).__init__()
        self.__evaluators = []

    def add(self, evaluator):
        """
        :param ContextualEvaluator evaluator:
        """
        self.__evaluators.append(evaluator)

    def evaluate(self, request, context):
        for evaluator in self.__evaluators:
            if evaluator.supports(request):
                return evaluator.evaluate(request, context)
        raise Exception('Unsupported Evaluator.Request [{}]'.format(request))



from unittest import TestCase
from unittest.mock import Mock

from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.evaluation.match.target_matcher import TargetMatcher
from hackle.internal.evaluation.target.experiment_target_determiner import ExperimentTargetDeterminer
from tests.internal.evaluation.evaluator.experiment.test_experiment_request import create_experiment_request
from tests.internal.model.test_entities import create_experiment


class ExperimentTargetDeterminerTest(TestCase):
    def setUp(self):
        self.target_matches = []
        self.target_matcher = Mock(spec=TargetMatcher)
        self.target_matcher.matches = Mock(side_effect=self.target_matches)
        self.sut = ExperimentTargetDeterminer(self.target_matcher)

    def test_audiences가_비어있으면_true(self):
        # given
        request = create_experiment_request()

        # when
        actual = self.sut.is_user_in_experiment_target(request, Evaluator.context())

        # then
        self.assertTrue(actual)

    def test_하나라도_일치하는_audience가_있는경우_true(self):
        # given
        experiment = create_experiment(target_audiences=[
            self._audience_target(False),
            self._audience_target(False),
            self._audience_target(False),
            self._audience_target(True),
            self._audience_target(False),
        ])
        request = create_experiment_request(experiment=experiment)

        # when
        actual = self.sut.is_user_in_experiment_target(request, Evaluator.context())

        # then
        self.assertTrue(actual)
        self.assertEqual(4, self.target_matcher.matches.call_count)

    def test_일치하는_audience가_하나도_없는경우_false(self):
        # given
        experiment = create_experiment(target_audiences=[
            self._audience_target(False),
            self._audience_target(False),
            self._audience_target(False),
            self._audience_target(False),
            self._audience_target(False),
        ])
        request = create_experiment_request(experiment=experiment)

        # when
        actual = self.sut.is_user_in_experiment_target(request, Evaluator.context())

        # then
        self.assertFalse(actual)
        self.assertEqual(5, self.target_matcher.matches.call_count)

    def _audience_target(self, is_match):
        self.target_matches.append(is_match)
        return Mock()



from threading import Thread


class Jobs:

    def __init__(self, count, task):
        self.__jobs = [Thread(target=task) for _ in range(count)]

    def start(self):
        for job in self.__jobs:
            job.start()

    def wait(self):
        for job in self.__jobs:
            job.join()

    def start_and_wait(self):
        self.start()
        self.wait()



from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.evaluation.evaluator.remoteconfig.remote_config_request import RemoteConfigRequest
from hackle.internal.model.properties_builder import PropertiesBuilder


class RemoteConfigEvaluation(Evaluator.Evaluation):
    def __init__(self, reason, target_evaluations, parameter, value_id, value, properties):
        """

        :param str reason:
        :param list[Evaluator.Evaluation] target_evaluations:
        :param RemoteConfigParameter parameter:
        :param int or None value_id:
        :param object or None value:
        :param dict[str, object] properties:
        """
        super(RemoteConfigEvaluation, self).__init__(reason, target_evaluations)
        self.__parameter = parameter
        self.__value_id = value_id
        self.__value = value
        self.__properties = properties

    @staticmethod
    def of(request, context, value_id, value, reason, properties_builder):
        """
        :param RemoteConfigRequest request:
        :param Evaluator.Context context:
        :param int or None value_id:
        :param object value:
        :param str reason:
        :param PropertiesBuilder properties_builder:

        :rtype: RemoteConfigEvaluation
        """
        properties_builder.add("returnValue", value)
        return RemoteConfigEvaluation(
            reason,
            context.target_evaluations,
            request.parameter,
            value_id,
            value,
            properties_builder.build()
        )

    @staticmethod
    def of_default(request, context, reason, properties_builder):
        """

        :param RemoteConfigRequest request:
        :param Evaluator.Context context:
        :param str reason:
        :param PropertiesBuilder properties_builder:

        :rtype: RemoteConfigEvaluation
        """
        return RemoteConfigEvaluation.of(request, context, None, request.default_value, reason, properties_builder)

    @property
    def parameter(self):
        """
        :rtype: RemoteConfigParameter
        """
        return self.__parameter

    @property
    def value_id(self):
        """
        :rtype: int or None
        """
        return self.__value_id

    @property
    def value(self):
        """
        :rtype: object or None
        """
        return self.__value

    @property
    def properties(self):
        """
        :rtype: dict[str, object]
        """
        return self.__properties



class InMemoryEventProcessor(object):
    def __init__(self):
        self.__events = []

    @property
    def processed_events(self):
        return list(self.__events)

    def process(self, event):
        self.__events.append(event)



from hackle.internal.evaluation.evaluator.evaluator import Evaluator
from hackle.internal.evaluation.evaluator.experiment.experiment_evaluation import ExperimentEvaluation
from hackle.internal.evaluation.evaluator.remoteconfig.remote_config_evaluation import RemoteConfigEvaluation
from hackle.internal.event.user_event import UserEvent
from hackle.internal.model.properties_builder import PropertiesBuilder
from hackle.internal.time.clock import SYSTEM_CLOCK


class UserEventFactory(object):
    __ROOT_TYPE = "$targetingRootType"
    __ROOT_ID = "$targetingRootId"
    __CONFIG_ID_PROPERTY_KEY = "$parameterConfigurationId"

    def __init__(self, clock=SYSTEM_CLOCK):
        self.__clock = clock

    def create(self, request, evaluation):
        """
        :param Evaluator.Request request:
        :param Evaluator.Evaluation evaluation:
        :rtype: list[UserEvent]
        """

        timestamp = self.__clock.current_millis()
        events = []

        root_event = self.__create(request, evaluation, timestamp, PropertiesBuilder())
        events.append(root_event)

        for target_evaluation in evaluation.target_evaluations:
            properties_builder = PropertiesBuilder()
            properties_builder.add(UserEventFactory.__ROOT_TYPE, request.key.type)
            properties_builder.add(UserEventFactory.__ROOT_ID, request.key.id)
            target_event = self.__create(request, target_evaluation, timestamp, properties_builder)
            events.append(target_event)

        return events

    # noinspection PyMethodMayBeStatic
    def __create(self, request, evaluation, timestamp, properties_builder):
        """

        :param Evaluator.Request request:
        :param Evaluator.Evaluation evaluation:
        :param int timestamp:
        :param PropertiesBuilder properties_builder:
        :return:
        """
        if isinstance(evaluation, ExperimentEvaluation):
            if evaluation.config is not None:
                properties_builder.add(UserEventFactory.__CONFIG_ID_PROPERTY_KEY, evaluation.config.id)
            return UserEvent.exposure(request.user, evaluation, properties_builder.build(), timestamp)

        if isinstance(evaluation, RemoteConfigEvaluation):
            properties_builder.add_properties(evaluation.properties)
            return UserEvent.remote_config(request.user, evaluation, properties_builder.build(), timestamp)

        raise Exception('Unsupported Evaluation [{}]'.format(evaluation.__class__.__name__))



from unittest import TestCase
from unittest.mock import Mock

from hackle.internal.evaluation.match.operator_matcher import OperatorMatcher, InMatcher
from hackle.internal.evaluation.match.value_matcher import StringValueMatcher, NumberValueMatcher, BoolValueMatcher, \
    VersionValueMatcher, NoneValueMatcher


class StringValueMatcherTest(TestCase):

    def setUp(self):
        self.sut = StringValueMatcher()

    def __check(self, user_value, match_value, is_matched):
        self.assertEqual(self.sut.matches(InMatcher(), user_value, match_value), is_matched)

    def test_string_type(self):
        self.__check("42", "42", True)

    def test_number_type(self):
        self.__check(42, 42, True)
        self.__check("42", 42, True)
        self.__check(42, "42", True)

        self.__check(42.42, 42.42, True)
        self.__check("42.42", 42.42, True)
        self.__check(42.42, "42.42", True)

        self.__check(42.0, 42.0, True)
        self.__check("42.0", 42.0, True)
        self.__check(42.0, "42.0", True)

    def test_unsupported_type(self):
        self.__check(True, True, False)
        self.__check(True, "True", False)
        self.__check("True", True, False)


class NumberValueMatcherTest(TestCase):

    def setUp(self):
        self.sut = NumberValueMatcher()

    def __check(self, user_value, match_value, is_matched):
        self.assertEqual(self.sut.matches(InMatcher(), user_value, match_value), is_matched)

    def test_number_type(self):
        self.__check(42, 42, True)
        self.__check(42.42, 42.42, True)
        self.__check(42.0, 42, True)
        self.__check(42, 42.0, True)
        self.__check(0, 0, True)

    def test_string_type(self):
        self.__check("42", "42", True)
        self.__check("42", 42, True)
        self.__check(42, "42", True)

        self.__check("42.42", "42.42", True)
        self.__check("42.42", 42.42, True)
        self.__check(42.42, "42.42", True)

        self.__check("42.0", "42.0", True)
        self.__check("42.0", 42.0, True)
        self.__check(42.0, "42.0", True)

    def test_unsupported_type(self):
        self.__check("42a", 42, False)
        self.__check(0, False, False)
        self.__check(True, True, False)


class BoolValueMatcherTest(TestCase):

    def setUp(self):
        self.sut = BoolValueMatcher()

    def test_user_value_match_value_둘다_Bool타입이면_operator_matcher의_일치_결과를_리턴(self):
        # given
        operator_matcher = Mock(spec=OperatorMatcher)
        operator_matcher.bool_matches.return_value = True

        # when
        actual = self.sut.matches(operator_matcher, True, True)

        # then
        self.assertTrue(actual)
        operator_matcher.bool_matches.assert_called_once_with(True, True)

    def test_user_value가_Bool_타입이_아니면_false(self):
        # given
        operator_matcher = Mock(spec=OperatorMatcher)

        # when
        actual = self.sut.matches(operator_matcher, 'True', True)

        # then
        self.assertFalse(actual)
        operator_matcher.assert_not_called()

    def test_user_value가_Bool타입이지만_match_value가_String타입이_아니면_false(self):
        # given
        operator_matcher = Mock(spec=OperatorMatcher)

        # when
        actual = self.sut.matches(operator_matcher, True, 'True')

        # then
        self.assertFalse(actual)
        operator_matcher.assert_not_called()


class VersionValueMatcherTest(TestCase):

    def setUp(self):
        self.sut = VersionValueMatcher()

    def test_user_value_match_value_둘다_version_타입이면_operator_matcher_결과를_리턴(self):
        # given
        operator_matcher = Mock(spec=OperatorMatcher)
        operator_matcher.version_matches.return_value = True

        # when
        actual = self.sut.matches(operator_matcher, '1.0.0', '2.0.0')

        # then
        self.assertTrue(actual)
        operator_matcher.version_matches.assert_called_once()

    def test_user_value가_version_타입이_아니면_False(self):
        # given
        operator_matcher = Mock(spec=OperatorMatcher)

        # when
        actual = self.sut.matches(operator_matcher, 1, '1.0.0')

        # then
        self.assertFalse(actual)
        operator_matcher.assert_not_called()

    def test_user_value가_version_타입이지만_match_value가_version_타입이_아니면_False(self):
        # given
        operator_matcher = Mock(spec=OperatorMatcher)

        # when
        actual = self.sut.matches(operator_matcher, '1.0.0', 1)

        # then
        self.assertFalse(actual)
        operator_matcher.assert_not_called()


class NoneValueMatcherTest(TestCase):

    def setUp(self):
        self.sut = NoneValueMatcher()

    def test_user_value_match_value_가_어떤_타입이건_False를_리턴(self):
        # given
        operator_matcher = Mock(spec=OperatorMatcher)

        # when
        # then
        self.assertFalse(self.sut.matches(operator_matcher, None, None))
        self.assertFalse(self.sut.matches(operator_matcher, None, 'test'))
        self.assertFalse(self.sut.matches(operator_matcher, 'test', 'test'))
        self.assertFalse(self.sut.matches(operator_matcher, 'test', 10))
        self.assertFalse(self.sut.matches(operator_matcher, 10, 10))



import time
from unittest import TestCase

from hackle.internal.time.clock import SYSTEM_CLOCK, SystemClock, Clock


class SystemClockTest(TestCase):
    def test_global(self):
        self.assertIsInstance(SYSTEM_CLOCK, SystemClock)

    def test_current_millis(self):
        sut = SystemClock()

        s = sut.current_millis()
        time.sleep(0.1)
        e = sut.current_millis()

        self.assertTrue(95 < e - s < 105)


class FixedClock(Clock):

    def __init__(self, current_millis):
        """
        :param int current_millis:
        """
        super(FixedClock, self).__init__()
        self.__current_millis = current_millis

    def current_millis(self):
        return self.__current_millis



from unittest.mock import Mock

from hackle.internal.evaluation.match.condition_matcher import ConditionMatcher
from hackle.internal.evaluation.match.condition_matcher_factory import ConditionMatcherFactory
from hackle.internal.evaluation.match.target_matcher import TargetMatcher
from hackle.internal.model.entities import Target, TargetCondition, TargetKey
from tests import base


class TargetMatcherTest(base.BaseTest):

    def setUp(self):
        self.match_values = []
        self.condition_matcher = Mock(spec=ConditionMatcher)
        self.condition_matcher.matches = Mock(side_effect=self.match_values)

        self.condition_matcher_factory = Mock(spec=ConditionMatcherFactory)
        self.condition_matcher_factory.get_condition_matcher_or_none.return_value = self.condition_matcher

        self.sut = TargetMatcher(self.condition_matcher_factory)

    def test_모든_조건이_일치하면_true(self):
        # given
        target = Target([
            self._condition(True),
            self._condition(True),
            self._condition(True),
            self._condition(True),
            self._condition(True),
            self._condition(True),
        ])

        # when
        actual = self.sut.matches(Mock(), Mock(), target)

        # then
        self.assertTrue(actual)
        self.assertEqual(6, self.condition_matcher.matches.call_count)

    def test_타겟의_조건중_하나라도_일치하지_않으면_false(self):
        # given
        target = Target([
            self._condition(True),
            self._condition(True),
            self._condition(True),
            self._condition(False),
            self._condition(True),
            self._condition(True),
        ])

        # when
        actual = self.sut.matches(Mock(), Mock(), target)

        # then
        self.assertFalse(actual)
        self.assertEqual(4, self.condition_matcher.matches.call_count)

    def _condition(self, is_match):
        self.match_values.append(is_match)
        return TargetCondition(key=TargetKey(type='USER_PROPERTY', name='age'), match=Mock())
